"""gustaf/gustaf/helpers/options.py

Classes to help organize options.
"""
from copy import deepcopy

from gustaf import settings


class Option:
    """
    Minimal Class to hold each options. Mainly to replace nested dict.
    """

    __slots__ = (
        "backend",
        "key",
        "description",
        "allowed_types",
    )

    def __init__(self, backend, key, description, allowed_types):
        self.backend = backend
        self.key = key
        self.description = description
        self.allowed_types = allowed_types

    def __repr__(self):
        specific = "\n".join(
            [
                "",
                self.key,
                "=" * len(self.key),
                "backend: " + self.backend,
                "description:",
                "  " + str(self.description),
                "allowed_types:",
                "  " + str(self.allowed_types),
                super().__repr__(),
                "",
            ]
        )
        return specific


# predefine recurring options
vedo_common_options = (
    Option(
        "vedo",
        "c",
        "Color in {rgb, RGB, str of (hex, name), int}",
        (str, tuple, list, int),
    ),
    Option("vedo", "alpha", "Transparency in range [0, 1].", (float, int)),
    Option(
        "vedo",
        "data_name",
        "Name of vertex_data to show. "
        "Object must have vertex_data with the same name.",
        (str,),
    ),
    Option("vedo", "vertex_ids", "Show ids of vertices", (bool,)),
    Option("vedo", "element_ids", "Show ids of elements", (bool,)),
    Option(
        "vedo",
        "lighting",
        "Lighting options {'default', 'metallic', 'plastic', 'shiny', "
        "'glossy', 'ambient', 'off'}",
        (str,),
    ),
    Option("vedo", "cmap", "Colormap for vertex_data plots.", (str,)),
    Option("vedo", "vmin", "Minimum value for cmap", (float, int)),
    Option("vedo", "vmax", "Maximum value for cmap", (float, int)),
    Option(
        "vedo",
        "cmap_alpha",
        "Colormap Transparency in range [0, 1].",
        (float, int),
    ),
    Option(
        "vedo",
        "scalarbar",
        "Scalarbar describing cmap. At least an empty dict or "
        "dict with following items are accepted: "
        "{title: str, pos: tuple, title_yoffset: int, font_size: int, "
        "nlabels: int, c: str, horizontal: bool, use_alpha: bool, "
        "label_format: str}. Setting bool will add a default scalarbar",
        (bool, dict),
    ),
    Option(
        "vedo",
        "arrow_data",
        "Name of vertex_data to plot as arrow. Corresponding data should be "
        "at least 2D. If you want more control over arrows, consider creating "
        "edges using gus.create.edges.from_data().",
        (str,),
    ),
    Option(
        "vedo",
        "arrow_data_scale",
        "Scaling factor for arrow data.",
        (float, int),
    ),
    Option(
        "vedo",
        "arrow_data_color",
        "Color for arrow data. Can be either cmap name or color. For "
        "cmap, colors are based on the size of the arrows.",
        (str, tuple, list, int),
    ),
    Option(
        "vedo",
        "axes",
        "Configure a specific axes with options. Expect dict(), but setting "
        "True will set a default axes. For full options, see "
        "https://vedo.embl.es/autodocs/content/vedo/addons.html"
        "#vedo.addons.Axes",
        (bool, dict),
    ),
)


def make_valid_options(*options):
    """
    Forms valid options. Should run only once during module loading.

    Parameters
    ----------
    *options: Option

    Returns
    -------
    valid_options: dict()
    """
    valid_options = dict()
    for opt in options:
        if not isinstance(opt, Option):
            raise TypeError("Please use `Option` to define options.")

        if not valid_options.get(opt.backend, False):
            valid_options[opt.backend] = dict()

        valid_options[opt.backend][opt.key] = opt

    return valid_options


class ShowOption:
    """
    Behaves similar to dict, but will only accept a set of options that's
    applicable to the helpee class. Intended use is to create a
    subclass that would define valid options for helpee.
    Options should be described by Option object.
    Helps all the way up to initializing backend showables up to their backend
    specific common routines. ShowOption and ShowManager in a sense.
    """

    __slots__ = ("_helpee", "_options", "_backend")

    _valid_options = dict()

    _helps = None

    def __init__(self, helpee):
        """
        Parameters
        ----------
        helpee: object
        """
        self._helpee = helpee
        if not type(helpee).__qualname__.startswith(self._helps):
            raise TypeError(
                f"This show option is for {self._helps}. "
                f"Given helpee is {type(helpee)}."
            )
        self._options = dict()
        self._backend = settings.VISUALIZATION_BACKEND

        # initialize backend specific option holder
        self._options[self._backend] = dict()

    def __repr__(self):
        """
        Friendly info prints for show option.

        Parameters
        ----------
        None

        Returns
        -------
        description: str
        """
        valid_and_current = list()
        for vo in self._valid_options[self._backend].values():
            valid = str(vo)
            current = ""
            if vo.key in self.keys():
                current = "current option: " + str(self[vo.key])
            valid_and_current.append(valid + current)

        header = [
            f"ShowOption for {self._helps}",
            f"Selected Backend: {self._backend}",
        ]
        return "\n".join(header + valid_and_current)

    def __setitem__(self, key, value):
        """
        Sets option after checking its validity.

        Parameters
        ----------
        key: str
        value: object

        Returns
        -------
        None
        """
        if key in self._valid_options[self._backend].keys():
            # valid type check
            if not isinstance(
                value, self._valid_options[self._backend][key].allowed_types
            ):
                raise TypeError(
                    f"{type(value)} is invalid value type for '{key}'. "
                    f"Details for '{key}':\n"
                    f"{self._valid_options[self._backend][key]}"
                )

            # types are valid. let's add
            self._options[self._backend][key] = value

        elif key.startswith("backend"):
            # special keyword.
            if not isinstance(value, str):
                raise TypeError(
                    f"Invalid backend info ({value}). Must be a str"
                )
            self._backend = value
            if not self._options.get(self._backend, False):
                self._options[self._backend] = dict()

        else:
            raise ValueError(f"{key} is an invalid option for {self._helps}.")

    def __getitem__(self, key):
        """
        operator[]

        Parameters
        ----------
        key: str or iterable

        Returns
        -------
        items: object or dict
        """
        if isinstance(key, str):
            return self._options[self._backend][key]
        elif hasattr(key, "__iter__"):
            items = dict()
            for k in key:
                if k in self._options[self._backend]:
                    items[k] = self._options[self._backend][k]
            return items
        else:
            raise TypeError(f"Invalid key type for {type(self)}")

    def get(self, key, default):
        """
        Gets value from key and default. Similar to dict.get()

        Parameters
        ----------
        key: stir
        default: object

        Returns
        -------
        values: object
        """
        return self._options[self._backend].get(key, default)

    def update(self, **kwargs):
        """
        Calls __setitem__ iteratively for validity check.

        Parameters
        ----------
        **kwargs: kwargs

        Returns
        -------
        None
        """
        for k, v in kwargs.items():
            self.__setitem__(k, v)

    def valid_keys(self, backend=None):
        """
        Returns valid keys. Can directly specify backend. If not, returns
        valid_keys for currently selected backend.

        Parameters
        ----------
        backend: str

        Returns
        -------
        valid_keys: dict_keys
        """
        backend = self._backend if backend is None else backend
        return self._valid_options[backend].keys()

    def keys(self):
        """
        Registered option keys.

        Parameters
        ----------
        None

        Returns
        -------
        keys: dict_keys
        """
        return self._options[self._backend].keys()

    def values(self):
        """
        Registered option values.

        Parameters
        ----------
        None

        Returns
        -------
        keys: dict_values
        """
        return self._options[self._backend].values()

    def items(self):
        """
        Registered option items.

        Parameters
        ----------
        None

        Returns
        -------
        items: dict_items
        """
        return self._options[self._backend].items()

    def clear(self):
        """
        Clears all the options.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        self._options.clear()
        # put back default backend option dict
        self._options[self._backend] = dict()

    def pop(self, *args, **kwargs):
        """
        Calls pop() on current backend options

        Parameters
        ----------
        None

        Returns
        -------
        value: object
        """
        self._options[self._backend].pop(*args, **kwargs)

    def copy_valid_options(self, copy_to, keys=None):
        """
        Copies valid option to other show_option. Simply iterates and tries.

        Parameters
        ----------
        copy_to: ShowOption
        keys: tuple or list
          Can specify keys

        Returns
        -------
        None
        """
        if not isinstance(copy_to, ShowOption):
            raise TypeError("copy_to should be a ShowOption")
        valid_keys = copy_to.valid_keys()

        if keys is not None:
            items = self[keys].items()
        else:
            items = self.items()

        for key, value in items:
            if key in valid_keys:
                copy_to[key] = deepcopy(value)  # is deepcopy necessary?

    def _initialize_showable(self):
        """
        Creates basic showable all the way up to backend common procedures.

        Parameters
        ----------
        None

        Returns
        -------
        showable: object
        """
        return eval(f"self._initialize_{self._backend}_showable()")



"""gustaf/spline/microstructure/__init__.py.

Interface for tools and generators creating simple microstructures.
"""

from gustaf.spline.microstructure import microstructure, tiles
from gustaf.spline.microstructure.microstructure import Microstructure

__all__ = ["tiles", "microstructure", "Microstructure"]



"""gustaf/gustaf/show.py.

Everything related to show/visualization.
"""
import sys

import numpy as np

from gustaf import settings, utils
from gustaf._base import GustafBase

# @linux it raises error if vedo is imported inside the function.
try:
    import vedo
except ImportError as err:
    # overwrites the vedo module with an object which will throw an error
    # as soon as it is used the first time. This means that any non vedo
    # functionality works as before, but as soon as vedo is used a
    # comprehensive exception will be raised which is understandable in
    # contrast to the possible errors previously possible
    from gustaf.helpers.raise_if import ModuleImportRaiser

    vedo = ModuleImportRaiser("vedo", err)


# enable `gus.show()`
# taken from https://stackoverflow.com/questions/1060796/callable-modules
# will use this until this module is renamed
class _CallableShowDotPy(sys.modules[__name__].__class__):
    def __call__(self, *args, **kwargs):
        """call show()"""
        return show(*args, **kwargs)


sys.modules[__name__].__class__ = _CallableShowDotPy


def show(*gus_obj, **kwargs):
    """Shows using appropriate backend.

    Parameters
    -----------
    *gus_obj: gustaf objects

    Returns
    --------
    None
    """
    vis_b = settings.VISUALIZATION_BACKEND

    if vis_b.startswith("vedo"):
        return show_vedo(*gus_obj, **kwargs)
    elif vis_b.startswith("trimesh"):
        pass
    elif vis_b.startswith("matplotlib"):
        pass
    else:
        raise NotImplementedError


def show_vedo(
    *args,
    **kwargs,
):
    """`vedo.show` wrapper. Each args represent one section of window. In other
    words len(args) == N, where N corresponds to the parameter for vedo.show().

    Parameters
    -----------
    *args: Union[List[Union[gustaf_obj, vedo_obj]], Dict[str, Any]]]
    """
    # vedo plotter parameter
    N = len(args)
    offs = kwargs.get("offscreen", False)
    interact = kwargs.get("interactive", True)
    plt = kwargs.get("vedoplot", None)
    skip_clear = kwargs.get("skip_clear", False)
    close = kwargs.get("close", None)
    size = kwargs.get("size", "auto")
    cam = kwargs.get("cam", None)
    title = kwargs.get("title", "gustaf")
    return_show_list = kwargs.get("return_showable_list", False)

    def clear_vedo_plotter(plotter, num_renderers, skip_cl=skip_clear):
        """enough said."""
        # for whatever reason it is desired
        if skip_cl:
            return None

        for i in range(num_renderers):
            plotter.clear(at=i)

        return None

    def cam_tuple_to_list(dict_cam):
        """if entity is tuple, turns it into list."""
        if dict_cam is None:
            return None

        for key, value in dict_cam.items():
            if isinstance(value, tuple):
                dict_cam[key] = list(value)

        return dict_cam

    # get plotter
    if plt is None:
        plt = vedo.Plotter(
            N=N, sharecam=False, offscreen=offs, size=size, title=title
        )

    else:
        # check if plt has enough Ns
        trueN = np.prod(plt.shape)
        clear_vedo_plotter(plt, trueN)  # always clear.
        if trueN != N:
            utils.log.warning(
                "Number of args exceed given vedo.Plotter's capacity.",
                "Assigning a new one",
            )
            title = plt.title
            if close:  # only if it is explicitly stated
                plt.close()  # Hope that this truly releases..
            # assign a new one
            plt = vedo.Plotter(
                N=N, sharecam=False, offscreen=offs, size=size, title=title
            )

    # loop and plot
    for i, arg in enumerate(args):
        # form valid input type.
        if isinstance(arg, dict):
            show_list = list(arg.values())
        elif isinstance(arg, list):
            show_list = arg.copy()
        else:
            # raise TypeError(
            #     "For vedo_show, only list or dict is valid input")
            utils.log.debug(
                "one of args for show_vedo is neither `dict` nor",
                "`list`. Putting it naively into a list.",
            )
            show_list = [arg]

        # quick check if the list is gustaf or non-gustaf
        # if gustaf, make it vedo-showable.
        # if there's spline, we need to pop the element and
        # extend showables to the list.
        # A show_list is a list to be plotted into a single sub frame of the
        # plot
        list_of_showables = []
        for sl in show_list:
            if not isinstance(sl, list):
                sl = [sl]
            for k, item in enumerate(sl):
                if isinstance(item, GustafBase):
                    tmp_showable = item.showable(backend="vedo", **kwargs)
                    # splines return dict
                    # - maybe it is time to do some typing..
                    if isinstance(tmp_showable, dict):
                        # add to extend later
                        list_of_showables.extend(list(tmp_showable.values()))
                    else:
                        # replace gustaf_obj with vedo_obj.
                        list_of_showables.append(tmp_showable)
                else:
                    list_of_showables.extend(sl)
        # set interactive to true at last element
        if int(i + 1) == len(args):
            plt.show(
                list_of_showables,
                at=i,
                interactive=interact,
                camera=cam_tuple_to_list(cam),
                # offscreen=offs,
            )

        else:
            plt.show(
                list_of_showables,
                at=i,
                interactive=False,
                camera=cam_tuple_to_list(cam),
                # offscreen=offs,
            )

    if interact and not offs:
        # only way to ensure memory is released
        clear_vedo_plotter(plt, np.prod(plt.shape))

        if close or close is None:  # explicitly given or None.
            # It seems to leak some memory, but here it goes.
            plt.close()  # if i close it, this cannot be reused...
            plt = None

    if return_show_list:
        return (plt, list_of_showables)
    else:
        return plt


def _vedo_showable(obj, as_dict=False, **kwargs):
    """Generates a vedo obj based on `kind` attribute from given obj, as well
    as show_options.

    Parameters
    -----------
    obj: gustaf obj
    as_dict: bool
      If True, returns vedo objects in a dict. Corresponding main objects will
      be available with ["main"] key. Else, returns vedo.Assembly object,
      where all the objects are grouped together.
    **kwargs: kwargs
      Will try to overwrite applicable items.

    Returns
    --------
    vedo_obj: vedo obj
    """
    # incase kwargs are defined, we will make a copy of the object and
    # try to overwrite all the applicable kwargs.
    if kwargs:
        # keep original ones and assign new show_options temporarily
        orig_show_options = obj.show_options
        obj._show_options = obj.__show_option__(obj)
        orig_show_options.copy_valid_options(obj.show_options)
        for key, value in kwargs.items():
            try:
                obj.show_options[key] = value
            except BaseException:
                utils.log.debug(
                    f"Skipping invalid option {key} for "
                    f"{obj.show_options._helps}"
                )
                continue

    # minimal-initialization of vedo objects
    vedo_obj = obj.show_options._initialize_showable()
    # as dict?
    if as_dict:
        return_as_dict = dict()

    # set common values. Could be a perfect place to try :=, but we want to
    # support p3.6.
    c = obj.show_options.get("c", None)
    if c is not None:
        vedo_obj.c(c)

    alpha = obj.show_options.get("alpha", None)
    if alpha is not None:
        vedo_obj.alpha(alpha)

    lighting = obj.show_options.get("lighting", None)
    if lighting is not None:
        vedo_obj.lighting(lighting)

    vertex_ids = obj.show_options.get("vertex_ids", False)
    element_ids = obj.show_options.get("element_ids", False)
    # special treatment for vertex
    if obj.kind.startswith("vertex"):
        vertex_ids = vertex_ids | element_ids
        if element_ids:
            utils.log.debug(
                "`element_ids` option is True for Vertices. Overwriting it as"
                "vertex_ids."
            )
            element_ids = False
    if vertex_ids:
        # use vtk font. supposedly faster. And differs from cell id.
        vertex_ids = vedo_obj.labels("id", on="points", font="VTK")
        if not as_dict:
            vedo_obj += vertex_ids
        else:
            return_as_dict["vertex_ids"] = vertex_ids
    if element_ids:
        # should only reach here if this obj is not vertex
        element_ids = vedo.Points(obj.centers()).labels("id", on="points")
        if not as_dict:
            vedo_obj += element_ids
        else:
            return_as_dict["element_ids"] = element_ids

    # data plotting
    data_name = obj.show_options.get("data_name", None)
    vertex_data = obj.vertex_data.as_scalar(data_name, None)
    if data_name is not None and vertex_data is not None:
        # transfer data
        if obj.kind.startswith("edge"):
            vedo_obj.pointdata[data_name] = vertex_data[obj.edges].reshape(
                -1, vertex_data.shape[1]
            )
        else:
            vedo_obj.pointdata[data_name] = vertex_data

        # form cmap kwargs for init
        cmap_keys = ("vmin", "vmax")
        cmap_kwargs = obj.show_options[cmap_keys]
        # set a default cmap if needed
        cmap_kwargs["input_cmap"] = obj.show_options.get("cmap", "plasma")
        cmap_kwargs["alpha"] = obj.show_options.get("cmap_alpha", 1)
        # add data_name
        cmap_kwargs["input_array"] = data_name

        # set cmap
        vedo_obj.cmap(**cmap_kwargs)

        # at last, scalarbar
        # deprecated function name, keep it for now for backward compat
        sb_kwargs = obj.show_options.get("scalarbar", None)
        if sb_kwargs is not None and sb_kwargs is not False:
            sb_kwargs = dict() if isinstance(sb_kwargs, bool) else sb_kwargs
            vedo_obj.addScalarBar(**sb_kwargs)

    elif data_name is not None and vertex_data is None:
        utils.log.debug(
            f"No vertex_data named '{data_name}' for {obj}. Skipping"
        )

    # arrow plots - this is independent from data plotting.
    arrow_data_name = obj.show_options.get("arrow_data", None)
    # will raise if data is scalar
    arrow_data_value = obj.vertex_data.as_arrow(arrow_data_name, None, True)
    if arrow_data_name is not None and arrow_data_value is not None:
        from gustaf.create.edges import from_data

        # we are here because this data is not a scalar
        # is showable?
        if arrow_data_value.shape[1] not in (2, 3):
            raise ValueError(
                "Only 2D or 3D data can be shown.",
                f"Requested data is {arrow_data_value.shape[1]}",
            )

        as_edges = from_data(
            obj,
            arrow_data_value,
            obj.show_options.get("arrow_data_scale", None),
            data_norm=obj.vertex_data.as_scalar(arrow_data_name),
        )
        arrows = vedo.Arrows(
            as_edges.vertices[as_edges.edges],
            c=obj.show_options.get("arrow_data_color", "plasma"),
        )
        if not as_dict:
            vedo_obj += arrows
        else:
            return_as_dict["arrow_data"] = arrows

    axes_kw = obj.show_options.get("axes", None)
    # need to explicitly check if it is false
    if axes_kw is not None and axes_kw is not False:
        axes_kw = dict() if isinstance(axes_kw, bool) else axes_kw
        axes = vedo.Axes(vedo_obj, **axes_kw)
        if not as_dict:
            vedo_obj += axes
        else:
            return_as_dict["axes"] = axes

    # set back temporary show_options if needed
    if kwargs:
        obj._show_options = orig_show_options

    if not as_dict:
        return vedo_obj
    else:
        return_as_dict["main"] = vedo_obj
        return return_as_dict


def _trimesh_showable(obj):
    """"""
    pass


def _matplotlib_showable(obj):
    """"""
    pass


def make_showable(obj, backend=settings.VISUALIZATION_BACKEND, **kwargs):
    """Since gustaf does not natively support visualization, one of the
    following library is used to visualize gustaf (visualizable) objects: (1)
    vedo -> Fast, offers a lot of features (2) trimesh -> Fast, compatible with
    old OpenGL (3) matplotlib -> Slow, offers vector graphics.

    This determines showing types using `whatami`.

    Parameters
    -----------
    obj: gustaf-objects
    backend: str
      (Optional) Default is `gustaf.settings.VISUALIZATION_BACKEND`.
      Options are: "vedo" | "trimesh" | "matplotlib"

    Returns
    --------
    showable_objs: list
      List of showable objects.
    """
    if backend.startswith("vedo"):
        return _vedo_showable(obj, **kwargs)
    elif backend.startswith("trimesh"):
        return _trimesh_showable(obj, **kwargs)
    elif backend.startswith("matplotlib"):
        return _matplotlib_showable(obj, **kwargs)
    else:
        raise NotImplementedError


# possibly relocate, is this actually used?
# could not find any usage in this repo
def interpolate_vedo_dictcam(cameras, resolutions, spline_degree=1):
    """Interpolate between vedo dict cameras.

    Parameters
    ------------
    cameras: list or tuple
    resolutions: int
    spline_degree: int
      if > 1 and splinepy is available and there are more than two cameras,
      we interpolate all the entries using spline.

    Returns
    --------
    interpolated_cams: list
    """
    try:
        import splinepy

        spp = True

    except ImportError:
        spp = False

    # quick type check loop
    cam_keys = ["pos", "focalPoint", "viewup", "distance", "clippingRange"]
    for cam in cameras:
        if not isinstance(cam, dict):
            raise TypeError("Only `dict` description of vedo cam is allowed.")
        else:
            for key in cam_keys:
                if cam[key] is None:
                    raise ValueError(
                        f"One of the camera does not contain `{key}` info"
                    )

    interpolated_cams = []
    total_cams = int(resolutions) * (len(cameras) - 1)

    if spp and spline_degree > 1 and len(cameras) > 2:
        if spline_degree > len(cameras):
            raise ValueError(
                "Not enough camera to interpolate with "
                f"spline degree {spline_degree}"
            )

        ps = []
        fs = []
        vs = []
        ds = []
        cs = []
        for cam in cameras:
            ps.append(list(cam[cam_keys[0]]))
            fs.append(list(cam[cam_keys[1]]))
            vs.append(list(cam[cam_keys[2]]))
            ds.append([float(cam[cam_keys[3]])])
            cs.append(list(cam[cam_keys[4]]))

        interpolated = dict()
        for i, prop in enumerate([ps, fs, vs, ds, cs]):
            i_spline = splinepy.BSpline()
            i_spline.interpolate_curve(
                query_points=prop,
                degree=spline_degree,
                save_query=False,
            )
            interpolated[cam_keys[i]] = i_spline.sample([total_cams])

        for i in range(total_cams):
            interpolated_cams.append(
                {
                    cam_keys[0]: interpolated[cam_keys[0]][i].tolist(),
                    cam_keys[1]: interpolated[cam_keys[1]][i].tolist(),
                    cam_keys[2]: interpolated[cam_keys[2]][i].tolist(),
                    cam_keys[3]: interpolated[cam_keys[3]][i][0],  # float?
                    cam_keys[4]: interpolated[cam_keys[4]][i].tolist(),
                }
            )

    else:
        i = 0
        for start_cam, end_cam in zip(cameras[:-1], cameras[1:]):
            if i == 0:
                interpolated = [
                    np.linspace(
                        start_cam[ckeys],
                        end_cam[ckeys],
                        resolutions,
                    ).tolist()
                    for ckeys in cam_keys
                ]

            else:
                interpolated = [
                    np.linspace(
                        start_cam[ckeys],
                        end_cam[ckeys],
                        int(resolutions + 1),
                    )[1:].tolist()
                    for ckeys in cam_keys
                ]

            i += 1

            for j in range(resolutions):
                interpolated_cams.append(
                    {
                        cam_keys[0]: interpolated[0][j],
                        cam_keys[1]: interpolated[1][j],
                        cam_keys[2]: interpolated[2][j],
                        cam_keys[3]: interpolated[3][j],  # float?
                        cam_keys[4]: interpolated[4][j],
                    }
                )

    return interpolated_cams



import itertools

import numpy as np

from gustaf._base import GustafBase
from gustaf.show import show_vedo
from gustaf.spline import base


class Microstructure(GustafBase):
    """Helper class to facilitatae the construction of microstructures."""

    def __init__(
        self,
        deformation_function=None,
        tiling=None,
        microtile=None,
        parametrization_function=None,
    ):
        """Helper class to facilitatae the construction of microstructures.

        Parameters
        ----------
        deformation_function : spline
          Outer function that describes the contour of the microstructured
          geometry
        tiling : list of integers
          microtiles per parametric dimension
        microtile : spline or list of splines
          Representation of the building block defined in the unit cube
        parametrization_function : Callable (optional)
          Function to describe spline parameters
        """
        if deformation_function is not None:
            self.deformation_function = deformation_function

        if tiling is not None:
            self.tiling = tiling

        if microtile is not None:
            self.microtile = microtile

        if parametrization_function is not None:
            self.parametrization_function = parametrization_function

    @property
    def deformation_function(self):
        """Deformation function defining the outer geometry (contour) of the
        microstructure.

        Parameters
        ----------
        None

        Returns
        -------
        deformation_function : spline
        """
        if hasattr(self, "_deformation_function"):
            return self._deformation_function
        else:
            return None

    @deformation_function.setter
    def deformation_function(self, deformation_function):
        """Deformation function setter defining the outer geometry of the
        microstructure. Must be spline type and as such inherit from
        gustaf.GustafSpline.

        Parameters
        ----------
        deformation_function : spline

        Returns
        -------
        None
        """
        if not issubclass(type(deformation_function), base.GustafSpline):
            raise ValueError(
                "Deformation function must be Gustaf-Spline."
                " e.g. gustaf.NURBS"
            )
        self._deformation_function = deformation_function
        self._sanity_check()

    @property
    def tiling(self):
        """Number of microtiles per parametric dimension.

        Parameters
        ----------
        None

        Returns
        -------
        tiling : list<int>
        """
        if hasattr(self, "_tiling"):
            return self._tiling
        else:
            return None

    @tiling.setter
    def tiling(self, tiling):
        """Setter for the tiling attribute, defining the number of microtiles
        per parametric dimension.

        Parameters
        ----------
        tiling : int / list<int>
          Number of tiles for each dimension respectively
        Returns
        -------
        None
        """
        if not isinstance(tiling, list):
            if not isinstance(tiling, int):
                raise ValueError(
                    "Tiling mus be either list of integers of integer " "value"
                )
        self._tiling = tiling
        # Is defaulted to False using function arguments
        self._sanity_check()
        self._logd(f"Successfully set tiling to : {self.tiling}")

    @property
    def microtile(self):
        """Microtile that is either a spline, a list of splines, or a class
        that provides a `create_tile` function."""
        if hasattr(self, "_microtile"):
            return self._microtile
        else:
            return None

    @microtile.setter
    def microtile(self, microtile):
        """Setter for microtile.

        Microtile must be either a spline, a list of splines, or a class that
        provides (at least) a `create_tile` function and a `dim` member.

        Parameters
        ----------
        microtile : spline / list<splines> / user-object
          arbitrary long list of splines that define the microtile

        Returns
        -------
        None
        """
        # place single tiles into a list to provide common interface
        if isinstance(microtile, list) or issubclass(
            type(microtile), base.GustafSpline
        ):
            microtile = self._make_microtilable(microtile)
        # Assign Microtile object to member variable
        self._microtile = microtile

        self._sanity_check()

    @property
    def parametrization_function(self):
        """Function, that - if required - parametrizes the microtiles.

        In order to use said function, the Microtile needs to provide a couple
        of attributes:

         - evaluation_points - a list of points defined in the unit cube
           that will be evaluated in the parametrization function to provide
           the required set of data points
         - parameter_space_dimension - dimensionality of the parametrization
           function and number of design variables for said microtile

        Parameters
        ----------
        None

        Returns
        -------
        parametrization_function : Callable
          Function that descibes the local tile parameters
        """
        if hasattr(self, "_parametrization_function"):
            return self._parametrization_function
        else:
            return None

    @parametrization_function.setter
    def parametrization_function(self, parametrization_function):
        if not callable(parametrization_function):
            raise TypeError("parametrization_function must be callable")
        self._parametrization_function = parametrization_function
        self._sanity_check()

    def create(self, closing_face=None, knot_span_wise=None, **kwargs):
        """Create a Microstructure.

        Parameters
        ----------
        closing_face : string
          If not None, Microtile must provide a function `closing_tile`
          Represents coordinate to be a closed surface {"x", "y", "z"}
        knot_span_wise : bool
          Insertion per knotspan vs. total number per paradim
        **kwargs
          will be passed to `create_tile` function

        Returns
        -------
        Microstructure : list<spline>
          finished microstructure based on object requirements
        """
        # Check if all information is gathered
        if not self._sanity_check():
            raise ValueError("Not enough information provided, abort")

        # Set default values
        if knot_span_wise is None:
            knot_span_wise = True

        # check if user wants closed structure
        closing_face_dim = {"x": 0, "y": 1, "z": 2}.get(closing_face)
        is_closed = closing_face_dim is not None
        if not is_closed and (closing_face is not None):
            raise ValueError(
                "Invalid format for closing_face argument, (handed: "
                f"{closing_face}), must be one of"
                "{'x', 'y', 'z'}"
            )

        if is_closed:
            is_closed = True
            if closing_face_dim >= self._deformation_function.para_dim:
                raise ValueError(
                    "closing face must be smaller than the deformation "
                    "function's parametric dimension"
                )
            if self._parametrization_function is None:
                raise ValueError(
                    "Faceclosure is currently only implemented for "
                    "parametrized microstructures"
                )

        # Prepare the deformation function
        # Transform into a non-uniform splinetype and make sure to work on copy
        if hasattr(self._deformation_function, "bspline"):
            deformation_function_copy = self._deformation_function.bspline
        else:
            deformation_function_copy = self._deformation_function.nurbs
        # Create Spline that will be used to iterate over parametric space
        ukvs = deformation_function_copy.unique_knots
        if knot_span_wise:
            for i_pd in range(deformation_function_copy.para_dim):
                if self.tiling[i_pd] == 1:
                    continue
                inv_t = 1 / self.tiling[i_pd]
                new_knots = [
                    j * inv_t * (ukvs[i_pd][i] - ukvs[i_pd][i - 1])
                    for i in range(1, len(ukvs[i_pd]))
                    for j in range(1, self.tiling[i_pd])
                ]
                # insert knots in both the deformation function
                deformation_function_copy.insert_knots(i_pd, new_knots)
        else:
            self._logd(
                "New knots will be inserted one by one with the objective"
                " to evenly distribute tiles within the parametric domain"
            )
            for i_pd in range(deformation_function_copy.para_dim):
                n_current_spans = len(ukvs[i_pd]) - 1
                if self.tiling[i_pd] == n_current_spans:
                    continue
                elif self.tiling[i_pd] < n_current_spans:
                    self._logw(
                        f"The requested tiling can not be provided, as "
                        f"there are too many knotspans in the deformation"
                        f" function. The tiling in parametric dimension "
                        f"{i_pd} will be set to {n_current_spans}"
                    )
                    self.tiling[i_pd] = n_current_spans
                else:
                    # Determine new knots
                    n_k_span = np.zeros(n_current_spans, dtype=int)
                    span_measure = np.diff(ukvs[i_pd])
                    for _ in range(self.tiling[i_pd] - n_current_spans):
                        add_knot = np.argmax(span_measure)
                        n_k_span[add_knot] += 1
                        span_measure[add_knot] *= n_k_span[add_knot] / (
                            n_k_span[add_knot] + 1
                        )

                    new_knots = []
                    for i, nks in enumerate(n_k_span):
                        new_knots.extend(
                            np.linspace(
                                ukvs[i_pd][i], ukvs[i_pd][i + 1], nks + 2
                            )[1:-1]
                        )
                    deformation_function_copy.insert_knots(i_pd, new_knots)

        # Bezier Extraction for composition
        def_fun_patches = deformation_function_copy.extract.beziers()

        # Calculate parametric space representation for parametrized
        # microstructures
        is_parametrized = self.parametrization_function is not None
        if is_parametrized:
            para_space_dimensions = [[u[0], u[-1]] for u in ukvs]
            def_fun_para_space = base.Bezier(
                degrees=[1] * deformation_function_copy.para_dim,
                control_points=np.array(
                    list(itertools.product(*para_space_dimensions[::-1]))
                )[:, ::-1],
            ).bspline
            for i_pd in range(deformation_function_copy.para_dim):
                if self.tiling[i_pd] != 1:
                    def_fun_para_space.insert_knots(
                        i_pd,
                        deformation_function_copy.unique_knots[i_pd][1:-1],
                    )
            def_fun_para_space = def_fun_para_space.extract.beziers()

        # Determine element resolution
        element_resolutions = [
            len(c) - 1 for c in deformation_function_copy.unique_knots
        ]

        # Start actual composition
        self._microstructure = []
        if is_parametrized:
            for i, (def_fun, def_fun_para) in enumerate(
                zip(def_fun_patches, def_fun_para_space)
            ):
                # Evaluate tile parameters
                positions = def_fun_para.evaluate(
                    self._microtile.evaluation_points
                )
                tile_parameters = self._parametrization_function(positions)

                # Check if center or closing tile
                if is_closed:
                    # check index
                    index = i
                    for ipd in range(closing_face_dim):
                        index -= index % element_resolutions[ipd]
                        index /= element_resolutions[ipd]
                    index = index % element_resolutions[closing_face_dim]
                    if index == 0:
                        # Closure at minimum id
                        tile = self._microtile.closing_tile(
                            parameters=tile_parameters,
                            closure=closing_face + "_min",
                            **kwargs,
                        )
                    elif (index + 1) == element_resolutions[closing_face_dim]:
                        # Closure at minimum id
                        tile = self._microtile.closing_tile(
                            parameters=tile_parameters,
                            closure=closing_face + "_max",
                            **kwargs,
                        )
                    else:
                        tile = self._microtile.create_tile(
                            parameters=tile_parameters, **kwargs
                        )
                else:
                    tile = self._microtile.create_tile(
                        parameters=tile_parameters, **kwargs
                    )

                # Perform composition
                for tile_patch in tile:
                    self._microstructure.append(def_fun.compose(tile_patch))
        # Not parametrized
        else:
            # Tile can be computed once (prevent to many evaluations)
            tile = self._microtile.create_tile(**kwargs)
            for def_fun in def_fun_patches:
                for t in tile:
                    self._microstructure.append(def_fun.compose(t))

        # return copy of precomputed member
        return self._microstructure.copy()

    def show(self, use_saved=False, return_gustaf=False, **kwargs):
        """
        Shows microstructure. Consists of deformation_function, microtile, and
        microstructure. Supported only by vedo.

        Parameters
        ----------
        use_saved: bool
        return_gustaf: bool
        **kwargs: kwargs
          Will be passed to show function

        Returns
        -------
        gustaf_obj: dict
          keys are deformation_function, microtile, and microstructure.
          Iff return_gustaf is True.
        plt: vedo.Plotter
        """
        if use_saved:
            if hasattr(self, "_microstructure"):
                microstructure = self._microstructure
            else:
                raise ValueError("No previous microstructure saved")
        else:
            # Create on the fly
            microstructure = self.create(**kwargs)

        # Precompute splines
        microtile = self.microtile.create_tile(**kwargs)
        deformation_function = self.deformation_function

        if return_gustaf:
            return dict(
                deformation_function=deformation_function,
                microtile=microtile,
                microstructure=microstructure,
            )

        # Show in vedo
        return show_vedo(
            ["Deformation Function", deformation_function],
            ["Microtile", microtile],
            ["Composed Microstructure", microstructure],
            **kwargs,
        )

    def _sanity_check(self):
        """Check all members and consistency of user data.

        Parameters
        ----------
        updated_properties : bool
          Sets the updated_properties variable to value, which indicates,
          wheither the microstructure needs to be rebuilt

        Returns
        -------
        passes: bool
        """
        if (
            (self.deformation_function is None)
            or (self.microtile is None)
            or (self.tiling is None)
        ):
            self._logd(
                "Current information not sufficient,"
                " awaiting further assignments"
            )
            return False
        # Check if microtile object fulfils requirements
        if not hasattr(self._microtile, "create_tile"):
            raise ValueError(
                "Microtile class does not provide the necessary "
                "attribute `create_tile`, that is required for "
                "microstructure construction"
            )
        if not hasattr(self._microtile, "dim"):
            raise ValueError(
                "Microtile class does not provide the necessary "
                "attribute `dim`, defining the dimensionality of "
                "the created tile"
            )

        # Check if parametric dimensions are consistent
        if not self.deformation_function.para_dim == self._microtile.dim:
            raise ValueError(
                "Microtile dimension must match parametric dimension of "
                "deformation function to enable composition"
            )

        # Check if tiling is consistent
        if isinstance(self.tiling, int):
            self.tiling = [self.tiling] * self.deformation_function.para_dim
        if len(self.tiling) != self.deformation_function.para_dim:
            raise ValueError(
                "Tiling list must have one entry per parametric dimension"
                " of the deformation function"
            )
        if self.parametrization_function is not None:
            self._logd("Checking compatibility of parametrization function")
            if not hasattr(self._microtile, "evaluation_points"):
                raise ValueError(
                    "Microtile class does not provide the necessary "
                    "attribute `evaluation_points`, that is required for"
                    " a parametrized microstructure construction"
                )
            if not hasattr(self._microtile, "parameter_space_dimension"):
                raise ValueError(
                    "Microtile class does not provide the necessary "
                    "attribute `parameter_space_dimension`, that is "
                    "required for a parametrized microstructure "
                    "construction"
                )
            result = self._parametrization_function(
                self._microtile.evaluation_points
            )
            if not isinstance(result, tuple):
                raise ValueError(
                    "Function outline of parametrization function must be "
                    "`f(np.ndarray)->tuple`"
                )
            if not len(result) == self._microtile.parameter_space_dimension:
                raise ValueError(
                    "Return type of Parametrization function is "
                    "insufficient, check documentation of Microtile for "
                    "dimensionality"
                )
        # Complete check
        return True

    def _make_microtilable(self, microtile):
        """Creates a Microtile object on the fly if user only provides (a list
        of) splines. Internal use only.

        Parameters
        ----------
        microtile : spline / list<splines>
          Microtile definition of a spline
        """
        return _UserTile(microtile)


class _UserTile:
    def __init__(self, microtile):
        """
        On the fly created class of a user tile
        Parameters
        ----------
        microtile : spline , list<splines>
        """
        # Assign microtiles
        self._user_tile = []

        if not isinstance(microtile, list):
            microtile = [microtile]

        for m in microtile:
            if not issubclass(type(m), base.GustafSpline):
                raise ValueError(
                    "Microtiles must be (list of) "
                    "gustaf.GustafSplines. e.g. gustaf.NURBS"
                )
            # Extract beziers for every non Bezier patch else this just
            # returns itself
            self._user_tile.extend(m.extract.beziers())
        self._dim = microtile[0].dim
        for m in microtile:
            if m.dim != self._dim:
                raise ValueError("Dimensions of spline lists inconsistent")

    @property
    def dim(self):
        return self._dim

    def create_tile(self, **kwargs):
        """Create a tile on the fly."""
        return self._user_tile.copy()



from gustaf.spline import base


class TileBase(base.GustafBase):
    """
    Base class for tile objects
    """

    def __init__(self):
        """
        Init Values to None
        """
        self._dim = None
        self._evaluation_points = None
        self._parameter_space_dimension = None

    @property
    def parameter_space_dimension(self):
        """Number of parameters per evaluation point."""
        if self._parameter_space_dimension is None:
            raise TypeError(
                "Inherited Tile-types need to provide "
                "_parameter_space_dimension, see documentation."
            )
        return self._parameter_space_dimension

    @property
    def evaluation_points(self):
        """Positions in the parametrization function to be evaluated when tile
        " "is constructed prior to composition.

        Parameters
        ----------
        None

        Returns
        -------
        evaluation_points : np.ndarray(6,3)
        """
        if self._evaluation_points is None:
            raise TypeError(
                "Inherited Tile-types need to provide _evaluation_points, see "
                "documentation."
            )
        return self._evaluation_points

    @property
    def dim(self):
        """Returns dimensionality in physical space of the Microtile.

        Parameters
        ----------
        None

        Returns
        -------
        dim : int
        """
        if self._dim is None:
            raise TypeError(
                "Inherited Tile-types need to provide _dim, see documentation."
            )
        return self._dim



import numpy as np

from gustaf.spline import base
from gustaf.spline.microstructure.tiles.tilebase import TileBase


class CrossTile3D(TileBase):
    def __init__(self):
        """Simple crosstile with linear-quadratic branches and a trilinear
        center spline."""
        self._dim = 3
        self._evaluation_points = np.array(
            [
                [0.0, 0.5, 0.5],
                [1.0, 0.5, 0.5],
                [0.5, 0.0, 0.5],
                [0.5, 1.0, 0.5],
                [0.5, 0.5, 0.0],
                [0.5, 0.5, 1.0],
            ]
        )
        self._parameter_space_dimension = 1

    def closing_tile(
        self,
        parameters=None,
        closure=None,
        boundary_width=0.1,
        filling_height=0.5,
        **kwargs,
    ):
        """Create a closing tile to match with closed surface.

        Parameters
        ----------
        parameters : tuple(np.ndarray)
          radii of fitting cylinder at evaluation points
        closure : str
          parametric dimension that needs to be closed.
          Must be {"z_min", "z_max"}
        boundary_width : float
          with of the boundary surronding branch
        filling_height : float
          portion of the height that is filled in parametric domain

        Returns
        -------
        list_of_splines : list
        """
        # Check parameters
        if closure is None:
            raise ValueError("No closing direction given")

        if parameters is None:
            self._logd("Tile request is not parametrized, setting default 0.2")
            parameters = tuple([np.ones(6) * 0.2])
        parameters = parameters[0]
        if not (np.all(parameters > 0) and np.all(parameters < 0.5)):
            raise ValueError("Thickness out of range (0, .5)")

        if not (0.0 < float(boundary_width) < 0.5):
            raise ValueError("Boundary Width is out of range")

        if not (0.0 < float(filling_height) < 1.0):
            raise ValueError("Filling must  be in (0,1)")

        inv_boundary_width = 1.0 - boundary_width
        inv_filling_height = 1.0 - filling_height
        center_width = 1.0 - 2 * boundary_width
        ctps_mid_height_top = (1 + filling_height) * 0.5
        ctps_mid_height_bottom = 1.0 - ctps_mid_height_top
        r_center = center_width * 0.5

        spline_list = []
        if closure == "z_min":
            # The branch is located at zmin of current tile
            branch_thickness = parameters[5]
            ctps_corner = np.array(
                [
                    [0.0, 0.0, 0.0],
                    [boundary_width, 0.0, 0.0],
                    [0.0, boundary_width, 0.0],
                    [boundary_width, boundary_width, 0.0],
                    [0.0, 0.0, filling_height],
                    [boundary_width, 0.0, filling_height],
                    [0.0, boundary_width, filling_height],
                    [boundary_width, boundary_width, filling_height],
                ]
            )

            spline_list.append(
                base.Bezier(degrees=[1, 1, 1], control_points=ctps_corner)
            )

            spline_list.append(
                base.Bezier(
                    degrees=[1, 1, 1],
                    control_points=(
                        ctps_corner + np.array([0.0, inv_boundary_width, 0.0])
                    ),
                )
            )

            spline_list.append(
                base.Bezier(
                    degrees=[1, 1, 1],
                    control_points=(
                        ctps_corner + np.array([inv_boundary_width, 0.0, 0.0])
                    ),
                )
            )

            spline_list.append(
                base.Bezier(
                    degrees=[1, 1, 1],
                    control_points=(
                        ctps_corner
                        + np.array(
                            [inv_boundary_width, inv_boundary_width, 0.0]
                        )
                    ),
                )
            )

            center_ctps = np.array(
                [
                    [boundary_width, boundary_width, 0.0],
                    [inv_boundary_width, boundary_width, 0.0],
                    [boundary_width, inv_boundary_width, 0.0],
                    [inv_boundary_width, inv_boundary_width, 0.0],
                    [boundary_width, boundary_width, filling_height],
                    [inv_boundary_width, boundary_width, filling_height],
                    [boundary_width, inv_boundary_width, filling_height],
                    [inv_boundary_width, inv_boundary_width, filling_height],
                ]
            )

            spline_list.append(
                base.Bezier(degrees=[1, 1, 1], control_points=center_ctps)
            )

            spline_list.append(
                base.Bezier(
                    degrees=[1, 1, 1],
                    control_points=np.maximum(
                        center_ctps - np.array([center_width, 0, 0]), 0
                    ),
                )
            )

            spline_list.append(
                base.Bezier(
                    degrees=[1, 1, 1],
                    control_points=np.maximum(
                        center_ctps - np.array([0, center_width, 0]), 0
                    ),
                )
            )

            spline_list.append(
                base.Bezier(
                    degrees=[1, 1, 1],
                    control_points=np.minimum(
                        center_ctps + np.array([center_width, 0, 0]), 1.0
                    ),
                )
            )

            spline_list.append(
                base.Bezier(
                    degrees=[1, 1, 1],
                    control_points=np.minimum(
                        center_ctps + np.array([0, center_width, 0]), 1.0
                    ),
                )
            )
            branch_ctps = np.array(
                [
                    [-r_center, -r_center, filling_height],
                    [r_center, -r_center, filling_height],
                    [-r_center, r_center, filling_height],
                    [r_center, r_center, filling_height],
                    [
                        -branch_thickness,
                        -branch_thickness,
                        ctps_mid_height_top,
                    ],
                    [branch_thickness, -branch_thickness, ctps_mid_height_top],
                    [-branch_thickness, branch_thickness, ctps_mid_height_top],
                    [branch_thickness, branch_thickness, ctps_mid_height_top],
                    [-branch_thickness, -branch_thickness, 1.0],
                    [branch_thickness, -branch_thickness, 1.0],
                    [-branch_thickness, branch_thickness, 1.0],
                    [branch_thickness, branch_thickness, 1.0],
                ]
            ) + np.array([0.5, 0.5, 0.0])

            spline_list.append(
                base.Bezier(degrees=[1, 1, 2], control_points=branch_ctps)
            )

            return spline_list
        elif closure == "z_max":
            # The branch is located at zmax of current tile
            branch_thickness = parameters[4]
            ctps_corner = np.array(
                [
                    [0.0, 0.0, inv_filling_height],
                    [boundary_width, 0.0, inv_filling_height],
                    [0.0, boundary_width, inv_filling_height],
                    [boundary_width, boundary_width, inv_filling_height],
                    [0.0, 0.0, 1.0],
                    [boundary_width, 0.0, 1.0],
                    [0.0, boundary_width, 1.0],
                    [boundary_width, boundary_width, 1.0],
                ]
            )

            spline_list.append(
                base.Bezier(degrees=[1, 1, 1], control_points=ctps_corner)
            )

            spline_list.append(
                base.Bezier(
                    degrees=[1, 1, 1],
                    control_points=(
                        ctps_corner + np.array([0.0, inv_boundary_width, 0.0])
                    ),
                )
            )

            spline_list.append(
                base.Bezier(
                    degrees=[1, 1, 1],
                    control_points=(
                        ctps_corner + np.array([inv_boundary_width, 0.0, 0.0])
                    ),
                )
            )

            spline_list.append(
                base.Bezier(
                    degrees=[1, 1, 1],
                    control_points=(
                        ctps_corner
                        + np.array(
                            [inv_boundary_width, inv_boundary_width, 0.0]
                        )
                    ),
                )
            )

            center_ctps = np.array(
                [
                    [boundary_width, boundary_width, inv_filling_height],
                    [inv_boundary_width, boundary_width, inv_filling_height],
                    [boundary_width, inv_boundary_width, inv_filling_height],
                    [
                        inv_boundary_width,
                        inv_boundary_width,
                        inv_filling_height,
                    ],
                    [boundary_width, boundary_width, 1.0],
                    [inv_boundary_width, boundary_width, 1.0],
                    [boundary_width, inv_boundary_width, 1.0],
                    [inv_boundary_width, inv_boundary_width, 1.0],
                ]
            )

            spline_list.append(
                base.Bezier(degrees=[1, 1, 1], control_points=center_ctps)
            )

            spline_list.append(
                base.Bezier(
                    degrees=[1, 1, 1],
                    control_points=np.maximum(
                        center_ctps - np.array([center_width, 0, 0]), 0
                    ),
                )
            )

            spline_list.append(
                base.Bezier(
                    degrees=[1, 1, 1],
                    control_points=np.maximum(
                        center_ctps - np.array([0, center_width, 0]), 0
                    ),
                )
            )

            spline_list.append(
                base.Bezier(
                    degrees=[1, 1, 1],
                    control_points=np.minimum(
                        center_ctps + np.array([center_width, 0, 0]), 1.0
                    ),
                )
            )

            spline_list.append(
                base.Bezier(
                    degrees=[1, 1, 1],
                    control_points=np.minimum(
                        center_ctps + np.array([0, center_width, 0]), 1.0
                    ),
                )
            )

            branch_ctps = np.array(
                [
                    [-branch_thickness, -branch_thickness, 0.0],
                    [branch_thickness, -branch_thickness, 0.0],
                    [-branch_thickness, branch_thickness, 0.0],
                    [branch_thickness, branch_thickness, 0.0],
                    [
                        -branch_thickness,
                        -branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [
                        branch_thickness,
                        -branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [
                        -branch_thickness,
                        branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [
                        branch_thickness,
                        branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [-r_center, -r_center, inv_filling_height],
                    [r_center, -r_center, inv_filling_height],
                    [-r_center, r_center, inv_filling_height],
                    [r_center, r_center, inv_filling_height],
                ]
            ) + np.array([0.5, 0.5, 0.0])

            spline_list.append(
                base.Bezier(degrees=[1, 1, 2], control_points=branch_ctps)
            )

            return spline_list
        else:
            raise NotImplementedError(
                "Requested closing dimension is not supported"
            )

    def create_tile(self, parameters=None, center_expansion=1.0, **kwargs):
        """Create a microtile based on the parameters that describe the branch
        thicknesses.

        Thickness parameters are used to describe the inner radius of the
        outward facing branches

        Parameters
        ----------
        parameters : tuple(np.array)
          only first entry is used, defines the internal radii of the
          branches
        center_expansion : float
          thickness of center is expanded by a factor

        Returns
        -------
        microtile_list : list(splines)
        """

        if not isinstance(center_expansion, float):
            raise ValueError("Invalid Type")
        if not ((center_expansion > 0.5) and (center_expansion < 1.5)):
            raise ValueError("Center Expansion must be in (.5,1.5)")
        max_radius = min(0.5, (0.5 / center_expansion))
        # set to default if nothing is given
        if parameters is None:
            self._logd("Setting branch thickness to default 0.2")
            parameters = tuple([np.ones(6) * 0.2])
        [x_min_r, x_max_r, y_min_r, y_max_r, z_min_r, z_max_r] = parameters[
            0
        ].tolist()
        for radius in [x_min_r, x_max_r, y_min_r, y_max_r, z_min_r, z_max_r]:
            if not isinstance(radius, float):
                raise ValueError("Invalid type")
            if not (radius > 0 and radius < max_radius):
                raise ValueError(
                    f"Radii must be in (0,{max_radius}) for "
                    f"center_expansion {center_expansion}"
                )

        # center radius
        center_r = (
            (x_min_r + x_max_r + y_min_r + y_max_r + z_min_r + z_max_r)
            / 6.0
            * center_expansion
        )
        hd_center = 0.5 * (0.5 + center_r)

        # Create the center-tile
        center_points = np.array(
            [
                [-center_r, -center_r, -center_r],
                [center_r, -center_r, -center_r],
                [-center_r, center_r, -center_r],
                [center_r, center_r, -center_r],
                [-center_r, -center_r, center_r],
                [center_r, -center_r, center_r],
                [-center_r, center_r, center_r],
                [center_r, center_r, center_r],
            ]
        )

        center_spline = base.Bezier(
            degrees=[1, 1, 1], control_points=center_points + [0.5, 0.5, 0.5]
        )

        # X-Axis branches
        # X-Min-Branch
        aux_x_min = min(x_min_r, center_r)
        x_min_ctps = np.array(
            [
                [-0.5, -x_min_r, -x_min_r],
                [-hd_center, -aux_x_min, -aux_x_min],
                center_points[0, :],
                [-0.5, x_min_r, -x_min_r],
                [-hd_center, aux_x_min, -aux_x_min],
                center_points[2, :],
                [-0.5, -x_min_r, x_min_r],
                [-hd_center, -aux_x_min, aux_x_min],
                center_points[4, :],
                [-0.5, x_min_r, x_min_r],
                [-hd_center, aux_x_min, aux_x_min],
                center_points[6, :],
            ]
        )
        x_min_spline = base.Bezier(
            degrees=[2, 1, 1], control_points=x_min_ctps + [0.5, 0.5, 0.5]
        )
        # X-Min-Branch
        aux_x_max = min(x_max_r, center_r)
        x_max_ctps = np.array(
            [
                center_points[1, :],
                [hd_center, -aux_x_max, -aux_x_max],
                [0.5, -x_max_r, -x_max_r],
                center_points[3, :],
                [hd_center, aux_x_max, -aux_x_max],
                [0.5, x_max_r, -x_max_r],
                center_points[5, :],
                [hd_center, -aux_x_max, aux_x_max],
                [0.5, -x_max_r, x_max_r],
                center_points[7, :],
                [hd_center, aux_x_max, aux_x_max],
                [0.5, x_max_r, x_max_r],
            ]
        )
        x_max_spline = base.Bezier(
            degrees=[2, 1, 1], control_points=x_max_ctps + [0.5, 0.5, 0.5]
        )

        # Y-Axis branches
        # Y-Min-Branch
        aux_y_min = min(y_min_r, center_r)
        y_min_ctps = np.array(
            [
                [-y_min_r, -0.5, -y_min_r],
                [y_min_r, -0.5, -y_min_r],
                [-aux_y_min, -hd_center, -aux_y_min],
                [aux_y_min, -hd_center, -aux_y_min],
                center_points[0, :],
                center_points[1, :],
                [-y_min_r, -0.5, y_min_r],
                [y_min_r, -0.5, y_min_r],
                [-aux_y_min, -hd_center, aux_y_min],
                [aux_y_min, -hd_center, aux_y_min],
                center_points[4, :],
                center_points[5, :],
            ]
        )
        y_min_spline = base.Bezier(
            degrees=[1, 2, 1], control_points=y_min_ctps + [0.5, 0.5, 0.5]
        )
        # Y-Min-Branch
        aux_y_max = min(y_max_r, center_r)
        y_max_ctps = np.array(
            [
                center_points[2, :],
                center_points[3, :],
                [-aux_y_max, hd_center, -aux_y_max],
                [aux_y_max, hd_center, -aux_y_max],
                [-y_max_r, 0.5, -y_max_r],
                [y_max_r, 0.5, -y_max_r],
                center_points[6, :],
                center_points[7, :],
                [-aux_y_max, hd_center, aux_y_max],
                [aux_y_max, hd_center, aux_y_max],
                [-y_max_r, 0.5, y_max_r],
                [y_max_r, 0.5, y_max_r],
            ]
        )
        y_max_spline = base.Bezier(
            degrees=[1, 2, 1], control_points=y_max_ctps + [0.5, 0.5, 0.5]
        )

        # Y-Axis branches
        # Y-Min-Branch
        aux_z_min = min(z_min_r, center_r)
        z_min_ctps = np.array(
            [
                [-z_min_r, -z_min_r, -0.5],
                [z_min_r, -z_min_r, -0.5],
                [-z_min_r, z_min_r, -0.5],
                [z_min_r, z_min_r, -0.5],
                [-aux_z_min, -aux_z_min, -hd_center],
                [aux_z_min, -aux_z_min, -hd_center],
                [-aux_z_min, aux_z_min, -hd_center],
                [aux_z_min, aux_z_min, -hd_center],
                center_points[0, :],
                center_points[1, :],
                center_points[2, :],
                center_points[3, :],
            ]
        )
        z_min_spline = base.Bezier(
            degrees=[1, 1, 2], control_points=z_min_ctps + [0.5, 0.5, 0.5]
        )
        # Y-Min-Branch
        aux_z_max = min(z_max_r, center_r)
        z_max_ctps = np.array(
            [
                center_points[4, :],
                center_points[5, :],
                center_points[6, :],
                center_points[7, :],
                [-aux_z_max, -aux_z_max, hd_center],
                [aux_z_max, -aux_z_max, hd_center],
                [-aux_z_max, aux_z_max, hd_center],
                [aux_z_max, aux_z_max, hd_center],
                [-z_max_r, -z_max_r, 0.5],
                [z_max_r, -z_max_r, 0.5],
                [-z_max_r, z_max_r, 0.5],
                [z_max_r, z_max_r, 0.5],
            ]
        )
        z_max_spline = base.Bezier(
            degrees=[1, 1, 2], control_points=z_max_ctps + [0.5, 0.5, 0.5]
        )

        return [
            center_spline,
            x_min_spline,
            x_max_spline,
            y_min_spline,
            y_max_spline,
            z_min_spline,
            z_max_spline,
        ]



from gustaf import (
    _version,
    create,
    edges,
    faces,
    helpers,
    io,
    settings,
    show,
    utils,
    vertices,
    volumes,
)
from gustaf.edges import Edges
from gustaf.faces import Faces
from gustaf.vertices import Vertices
from gustaf.volumes import Volumes

has_spline = False
try:
    from gustaf import spline
    from gustaf.spline.base import NURBS, Bezier, BSpline, RationalBezier
    from gustaf.spline.ffd import FFD

    has_spline = True
except ImportError as err:
    # overwrites the all modules which depend on the `splinepy` library
    # with an object which will throw an error
    # as soon as it is used the first time. This means that any non spline
    # functionality works as before, but as soon as these are used a
    # comprehensive exception will be raised which is understandable in
    # contrast to the possible multitude of errors previously possible
    from gustaf.helpers.raise_if import ModuleImportRaiser

    spline = ModuleImportRaiser("splinepy", err)
    BSpline = spline
    NURBS = spline
    Bezier = spline
    RationalBezier = spline
    FFD = spline


__version__ = _version.version

__all__ = [
    "__version__",
    "settings",
    "vertices",
    "edges",
    "faces",
    "volumes",
    "show",
    "utils",
    "create",
    "io",
    "helpers",
    "Vertices",
    "Edges",
    "Faces",
    "Volumes",
    "spline",
    "has_spline",
    "BSpline",
    "NURBS",
    "Bezier",
    "RationalBezier",
    "FFD",
]



"""gustaf/gustaf/helpers/data.py.

Helps helpee to manage data. Some useful data structures.
"""


from collections import namedtuple
from functools import wraps

import numpy as np

from gustaf._base import GustafBase


class TrackedArray(np.ndarray):
    """Taken from nice implementations of `trimesh` (see LICENSE.txt).
    `https://github.com/mikedh/trimesh/blob/main/trimesh/caching.py`. Minor
    adaption, since we don't have hashing functionalities.

    All the inplace functions will set modified flag and if some operations
    has potential to cause un-trackable behavior, writeable flags will be set
    to False.

    Note, if you really really want, it is possible to change the tracked
    array without setting modified flag.
    """

    __slots__ = ("_modified", "_source")

    def __array_finalize__(self, obj):
        """Sets default flags for any arrays that maybe generated based on
        tracked array."""
        self._modified = True
        self._source = int(0)

        if isinstance(obj, type(self)):
            if isinstance(obj._source, int):
                self._source = obj
            else:
                self._source = obj._source

    @property
    def mutable(self):
        return self.flags["WRITEABLE"]

    @mutable.setter
    def mutable(self, value):
        self.flags.writeable = value

    def _set_modified(self):
        """set modified flags to itself and to the source."""
        self._modified = True
        if isinstance(self._source, type(self)):
            self._source._modified = True

    def copy(self, *args, **kwargs):
        """copy gives np.ndarray.

        no more tracking.
        """
        return np.array(self, copy=True)

    def view(self, *args, **kwargs):
        """Set writeable flags to False for the view."""
        v = super(self.__class__, self).view(*args, **kwargs)
        v.flags.writeable = False
        return v

    def __iadd__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__iadd__(*args, **kwargs)

    def __isub__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__isub__(*args, **kwargs)

    def __imul__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__imul__(*args, **kwargs)

    def __idiv__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__idiv__(*args, **kwargs)

    def __itruediv__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__itruediv__(*args, **kwargs)

    def __imatmul__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__imatmul__(*args, **kwargs)

    def __ipow__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__ipow__(*args, **kwargs)

    def __imod__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__imod__(*args, **kwargs)

    def __ifloordiv__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__ifloordiv__(*args, **kwargs)

    def __ilshift__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__ilshift__(*args, **kwargs)

    def __irshift__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__irshift__(*args, **kwargs)

    def __iand__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__iand__(*args, **kwargs)

    def __ixor__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__ixor__(*args, **kwargs)

    def __ior__(self, *args, **kwargs):
        self._set_modified()
        return super(self.__class__, self).__ior__(*args, **kwargs)

    def __setitem__(self, *args, **kwargs):
        self._set_modified()
        super(self.__class__, self).__setitem__(*args, **kwargs)

    def __setslice__(self, *args, **kwargs):
        self._set_modified()
        super(self.__class__, self).__setslice__(*args, **kwargs)

    def __getslice__(self, *args, **kwargs):
        self._set_modified()
        """
        return slices I am pretty sure np.ndarray does not have __*slice__
        """
        slices = super(self.__class__, self).__getitem__(*args, **kwargs)
        if isinstance(slices, np.ndarray):
            slices.flags.writeable = False
        return slices


def make_tracked_array(array, dtype=None, copy=True):
    """Taken from nice implementations of `trimesh` (see LICENSE.txt).
    `https://github.com/mikedh/trimesh/blob/main/trimesh/caching.py`.

    ``Properly subclass a numpy ndarray to track changes.
    Avoids some pitfalls of subclassing by forcing contiguous
    arrays and does a view into a TrackedArray.``

    Factory-like wrapper function for TrackedArray.

    Parameters
    ------------
    array: array- like object
      To be turned into a TrackedArray
    dtype: np.dtype
      Which dtype to use for the array
    copy: bool
      Default is True. copy if True.

    Returns
    ------------
    tracked : TrackedArray
      Contains input array data
    """
    # if someone passed us None, just create an empty array
    if array is None:
        array = []
    # make sure it is contiguous then view it as our subclass
    tracked = np.ascontiguousarray(array, dtype=dtype)
    if copy:
        tracked = tracked.copy().view(TrackedArray)
    else:
        tracked = tracked.view(TrackedArray)

    # should always be contiguous here
    assert tracked.flags["C_CONTIGUOUS"]

    return tracked


class DataHolder(GustafBase):
    __slots__ = (
        "_helpee",
        "_saved",
    )

    def __init__(self, helpee):
        """Base class for any data holder. Behaves similar to dict.

        Parameters
        -----------
        helpee: object
          GustafBase objects would probably make the most sense here.
        """
        self._helpee = helpee
        self._saved = dict()

    def __setitem__(self, key, value):
        """Raise Error to disable direct value setting.

        Parameters
        -----------
        key: str
        value: object
        """
        raise NotImplementedError(
            "Sorry, you can't set items directly for "
            f"{type(self).__qualname__}"
        )

    def __getitem__(self, key):
        """Returns stored item if the key exists.

        Parameters
        -----------
        key: str

        Returns
        --------
        value: object
        """
        if key in self._saved.keys():
            return self._saved[key]

        else:
            raise KeyError(f"`{key}` is not stored for {type(self._helpee)}")

    def pop(self, key, default=None):
        """
        Applied pop() to saved data

        Parameters
        ----------
        key: str
        default: object

        Returns
        -------
        value: object
        """
        return self._saved.pop(key, default)

    def clear(self):
        """
        Clears saved data by reassigning new dict
        """
        self._saved = dict()

    def get(self, key, default_values=None):
        """Returns stored item if the key exists. Else, given default value. If
        the key exist, default value always exists, since it is initialized
        that way.

        Parameters
        -----------
        key: str
        default_values: object

        Returns
        --------
        value: object
        """
        if key in self._saved.keys():
            return self._saved[key]
        else:
            return default_values

    def keys(self):
        """Returns keys of data holding dict.

        Returns
        --------
        keys: dict_keys
        """
        return self._saved.keys()

    def values(self):
        """Returns values of data holding dict.

        Returns
        --------
        values: dict_values
        """
        return self._saved.values()

    def items(self):
        """Returns items of data holding dict.

        Returns
        --------
        values: dict_values
        """
        return self._saved.items()


class ComputedData(DataHolder):
    _depends = None
    _inv_depends = None

    __slots__ = ()

    def __init__(self, helpee, **kwargs):
        """Stores last computed values.

        Keys are expected to be the same as helpee's function that computes the
         value.

        Parameters
        -----------
        helpee: GustafBase
        """
        super().__init__(helpee)

    @classmethod
    def depends_on(cls, var_names, make_property=False):
        """Decorator as classmethod.

        checks if the key should be computed. Two cases, where the answer is
        yes:

        1. there's modification on arrays that the key depend on.
            ->erases all other
        2. is corresponding value None?

        Supports multi-dependency

        Parameters
        -----------
        var_name: list
        make_property:
        """

        def inner(func):
            # followings are done once while modules are loaded
            # just subclass this class to make a special helper
            # for each helpee class.
            assert isinstance(var_names, list), "var_names should be a list"
            # initialize property
            # _depends is dict(str: list)
            if cls._depends is None:
                cls._depends = dict()
            if cls._depends.get(func.__name__, None) is None:
                cls._depends[func.__name__] = list()
            # add dependency info
            cls._depends[func.__name__].extend(var_names)

            # _inv_depends is dict(str: list)
            if cls._inv_depends is None:
                cls._inv_depends = dict()
            # add inverse dependency
            for vn in var_names:
                if cls._inv_depends.get(vn, None) is None:
                    cls._inv_depends[vn] = list()

                cls._inv_depends[vn].append(func.__name__)

            @wraps(func)
            def compute_or_return_saved(*args, **kwargs):
                """Check if the key should be computed,"""
                # extract some related info
                self = args[0]  # the helpee itself
                recompute = kwargs.get("recompute", False)
                return_saved = kwargs.get("return_saved", False)

                # if return_saved, try to escape as soon as possible
                if return_saved:
                    saved = self._computed._saved.get(func.__name__, None)
                    if saved is not None and not recompute:
                        return saved

                # computed arrays are called _computed.
                # loop over dependees and check if they are modified
                for dependee_str in cls._depends[func.__name__]:
                    dependee = getattr(self, dependee_str)
                    # is modified?
                    if dependee._modified:
                        for inv in cls._inv_depends[dependee_str]:
                            self._computed._saved[inv] = None

                # is saved / want to recompute?
                # recompute is added for computed values that accepts params.
                saved = self._computed._saved.get(func.__name__, None)
                if saved is not None and not recompute:
                    return saved

                # we've reached this point because we have to compute this
                computed = func(*args, **kwargs)
                if isinstance(computed, np.ndarray):
                    computed.flags.writeable = False  # configurable?
                self._computed._saved[func.__name__] = computed

                # so, all fresh. we can press NOT-modified  button
                for dependee_str in cls._depends[func.__name__]:
                    dependee = getattr(self, dependee_str)
                    dependee._modified = False

                return computed

            if make_property:
                return property(compute_or_return_saved)
            else:
                return compute_or_return_saved

        return inner


class VertexData(DataHolder):
    """
    Minimal manager for vertex data. Checks input array size, transforms
    data on request. __setitem__ and __getitem__ will perform length checks.
    key(), values(), items(), and get() will return whatever is currently
    stored.

    gustaf supports two kinds of data representation: scalar-data with cmap
    and vector-data with arrows.
    """

    __slots__ = ()

    def __init__(self, helpee):
        """Checks if helpee has vertices as attr beforehand.

        Parameters
        ----------
        helpee: Vertices
          Vertices and its derived classes.
        """
        if not hasattr(helpee, "vertices"):
            raise AttributeError("Helpee does not have `vertices`.")

        super().__init__(helpee)

    def _validate_len(self, value=None, raise_=True):
        """Checks if given value is a valid vertex_data based of its length.

        If raise_, throws error, else, deletes all incompatible values.
        Only checks len(). If array has (1, len) shape, this will still return
        False.

        Parameters
        ----------
        value: array-like
          Default is None. If None, checks all existing values.
        raise_: bool
          Default is True, If True, raises in case of incompatibility.

        Returns
        -------
        validity: bool
          If raise_ is False.
        """
        valid = True
        helpee_len = len(self._helpee.vertices)
        if value is not None:
            if len(value) != helpee_len:
                valid = False

            if raise_ and not valid:
                raise ValueError(
                    f"Expected ({helpee_len}) length data, "
                    f"Given (len(value))"
                )

            return valid

        # here, check all saved values.
        for key, value in self._saved.items():
            if len(value) != helpee_len:
                valid = False

            if not valid:
                if raise_:
                    raise ValueError(
                        f"`{key}`-data len ({len(value)}) doesn't match "
                        f"expected len ({helpee_len})"
                    )
                else:
                    self._logd(
                        f"`{key}`-data len ({len(value)}) doesn't match "
                        f"expect len ({helpee_len}). Deleting `{key}`."
                    )
                # pop invalid data
                self._saved.pop(key)
                self._saved.pop(key + "__norm", None)

        return valid

    def __setitem__(self, key, value):
        """
        Performs len() based check before storing vertex_data.

        Parameters
        ----------
        key: str
        value: object

        Returns
        -------
        None
        """
        self._validate_len(value, raise_=True)

        # we are here because this is valid
        self._saved[key] = make_tracked_array(value, copy=False).reshape(
            len(self._helpee.vertices), -1
        )

    def __getitem__(self, key):
        """
        Validates data length before returning item.

        Parameters
        ----------
        key: str

        Returns
        -------
        data: array-like
        """
        value = super().__getitem__(key)  # raises KeyError
        valid = self._validate_len(value, raise_=False)
        if valid:
            return value
        else:
            raise KeyError(
                "Either requested data is not stored or deleted due to "
                "changes in number of vertices."
            )

    def as_scalar(self, key, default=None):
        """
        Returns scalar version of requested data. If it is already a scalar,
        will return as is. Else, will return a norm. using `np.linalg.norm()`.

        Parameters
        ----------
        key: str
        default: object

        Returns
        -------
        data_as_scalar: (n, 1) np.ndarray
        """
        if key not in self.keys():
            return default

        # interpret scalar as norm
        # save the norm once it is called.
        if "__norm" not in key:
            norm_key = key + "__norm"
        else:
            norm_key = key
            key = key.replace("__norm", "")

        if norm_key in self.keys():
            saved = self[norm_key]  # performs len check
            # return if original is not modified
            if not self[key]._modified:  # check if original data is modified
                return saved
            else:
                self._saved.pop(norm_key)

        # we are here because we have to compute norm. let's save norm
        value = self[key]
        if value.shape[1] == 1:
            value_norm = value
        else:
            value_norm = np.linalg.norm(value, axis=1).reshape(-1, 1)

        # save norm
        self[norm_key] = value_norm
        # considered not modified
        self[key]._modified = False

        return value_norm

    def as_arrow(self, key, default=None, raise_=True):
        """
        Returns an array as is, only if it is showable as arrow.

        Parameters
        ----------
        key: str
        default: object
        raise_: bool

        Returns
        -------
        data: (n, d) np.ndarray
        """
        if key not in self.keys():
            return default

        value = self[key]
        if value.shape[1] == 1:
            self._logd(f"as_arrow() requested data ({key}) is 1D data.")
            if raise_:
                raise ValueError(
                    f"`{key}`-data is 1D and cannot be represented as arrows."
                )

        return value


class SplineDataAdaptor(GustafBase):
    """
    Prepares data to be presentable on spline. To support both
    scalar-data and vector-data, which are representable with colors and
    arrows respectively, this class will prepare data accordingly.
    """

    __slots__ = (
        "data",
        "function",
        "locations",
        "is_spline",
        "has_function",
        "has_locations",
        "has_evaluate",
        "arrow_data_only",
        "_user_created",
    )

    def __init__(self, data, locations=None, function=None):
        """ """
        # default
        self._user_created = True
        self.data = data
        self.is_spline = False
        self.has_function = False
        self.has_locations = False
        self.has_evaluate = False
        self.arrow_data_only = False

        # is spline we know?
        if "CoreSpline" in str(type(data).__mro__):
            self.is_spline = True

        # data has evaluate?
        if hasattr(data, "evaluate"):
            self.has_evaluate = callable(data.evaluate)

        # has function?
        if function is not None:
            self.has_function = True
            if not callable(function):
                raise ValueError("Given function isn't callable")
            self.function = function

        # locations? - keep this compatible with functions. maybe
        # we want to have some state dependent value at certain locations
        if locations is not None:
            # set what holds true
            self.has_locations = True
            self.arrow_data_only = True
            self.locations = np.asanyarray(locations)

            # if this is not a spline we know, it doesn't have a function,
            # it should:
            # -> `data.evaluate` is callable, or
            # -> len(data) == len(locations)
            if not self.is_spline and not self.has_function:
                len_matches = False
                if hasattr(data, "__len__"):
                    len_matches = len(locations) == len(data)
                if not (self.has_evaluate or len_matches):
                    raise ValueError(
                        "Data cannot be represented at specified locations."
                        "Requires one of the following requirements: "
                        "1) is a spline derived from splinepy's spline; "
                        "2) data has `data.evaluate()`; "
                        "3) length of the data and location should match."
                    )
            # location is specified, meaning we don't need sample()
            return None

        # can call sample or has a function?
        if not self.has_function and not self.is_spline:
            raise ValueError(
                "None spline data should at least have an accompanying "
                "function."
            )

    def as_vertex_data(self, resolutions=None, on=None):
        """
        Parameters
        ----------
        resolutions: list or tuple
        at: (n, d) array-like

        Returns
        -------
        vertex_data: (m, r) array-like
        """
        if resolutions is not None and on is not None:
            raise ValueError(
                "Please only specify either `resolutions` or `on`"
            )

        if self.has_locations and (resolutions is not None or on is not None):
            raise ValueError(
                "Location dependent data can't be evaluated with `resolutions`"
                " or `at`."
            )

        # if resolutions is specified, this is not a location query
        if resolutions is not None:
            if self.has_function:
                return self.function(self.data, resolutions=resolutions)
            elif self.is_spline and self.data.para_dim > 2:
                # TODO: replace this with generalized query helpers.
                return self.data.extract.faces(resolutions).vertices
            else:
                return self.data.sample(resolutions)

        # runtime location query
        if on is not None:
            if self.has_function:
                return self.function(self.data, on=on)
            elif self.has_evaluate:
                return self.data.evaluate(on)
            else:
                raise ValueError(
                    "Given data can't support data extraction on specified "
                    f"locations ({on})."
                )

        # location specified - either evaluate function at the
        if self.has_locations:
            if self.has_function:
                # function may want locations
                try:
                    return self.function(self.data, self.locations)
                except TypeError:  # maybe too many args
                    return self.function(self.data)
            elif self.has_evaluate:
                return self.data.evaluate(self.locations)
            else:
                return self.data

        # should be returned by now
        raise RuntimeError("Something went wrong while preparing spline data.")


class SplineData(DataHolder):
    """
    Data manager for splines.
    """

    def __init__(self, helpee):
        """ """
        if "GustafSpline" not in str(type(helpee).__mro__):
            raise AttributeError("Helpee does not have `vertices`")

        super().__init__(helpee)

    def __setitem__(self, key, value):
        """
        Selectively accept spline data.

        Parameters
        ----------
        key: str
        value: object

        Returns
        -------
        None
        """
        if isinstance(value, SplineDataAdaptor):
            self._saved[key] = value
        else:
            adapted = SplineDataAdaptor(value)  # will test usability
            adapted._user_created = False  # mark for __getitem__
            self._saved[key] = adapted

    def __getitem__(self, key):
        """
        Returns value from __setitem__

        Parameters
        ----------
        key: str

        Returns
        -------
        value: object
        """
        saved = super().__getitem__(key)
        if saved._user_created:
            return saved
        else:
            return saved.data

    def as_scalar(self, key, resolutions, default=None):
        """
        Return scalar value at given resolutions

        Parameters
        ----------
        key: str
        resolutions: list or tuple
        default: object
          Default is None and will return is key doesn't exist

        Returns
        -------
        value: np.ndarray
        """
        if key not in self._saved:
            return default

        saved = super().__getitem__(key)
        # will raise
        return saved.as_vertex_data(resolutions=resolutions)

    def as_arrow(self, key, resolutions=None, on=None, default=None):
        """
        Returns as-arrow-representable data on certain places, with given
        resolution, or on predefined places.

        Parameters
        ----------
        key: str
        resolutions: list or tuple
        on: array-like
        """
        if key not in self._saved:
            return default

        saved = super().__getitem__(key)
        # will raise
        return saved.as_vertex_data(resolutions=resolutions, on=on)


Unique2DFloats = namedtuple(
    "Unique2DFloats", ["values", "ids", "inverse", "intersection"]
)
Unique2DFloats.__doc__ = """
namedtuple to hold unique information of float type arrays.
Note that for float types, "close enough" might be a better name than unique.
This way, all tracked arrays, as long as they are 2D, have a dot separated
syntax to access unique info. For example, `mesh.unique_vertices.ids`.
"""
Unique2DFloats.values.__doc__ = """`(n, d) np.ndarray`
    Field number 0"""
Unique2DFloats.ids.__doc__ = """`(n, d) np.ndarray`
    Field number 1"""
Unique2DFloats.inverse.__doc__ = """`(n, d) np.ndarray`
    Field number 2"""
Unique2DFloats.intersection.__doc__ = """`(m) list of list`
  given original array's index, returns overlapping arrays, including itself.
  Field number 3
"""

Unique2DIntegers = namedtuple(
    "Unique2DIntegers", ["values", "ids", "inverse", "counts"]
)
Unique2DIntegers.__doc__ = """
namedtuple to hold unique information of integer type arrays.
Similar approach to Unique2DFloats.
"""

Unique2DIntegers.values.__doc__ = """`(n, d) np.ndarray`
    Field number 0"""
Unique2DIntegers.ids.__doc__ = """`(n) np.ndarray`
    Field number 1"""
Unique2DIntegers.inverse.__doc__ = """`(m) np.ndarray`
    Field number 2"""
Unique2DIntegers.counts.__doc__ = """`(n) np.ndarray`
    Field number 3"""


class ComputedMeshData(ComputedData):
    """A class to hold computed-mesh-data.

    Subclassed to keep its own dependency info.
    """

    pass



"""gustaf/spline/create.py.

Creates splines.
"""

import numpy as np

from gustaf import settings, utils


def extruded(spline, extrusion_vector=None):
    """Extrudes Splines.

    Parameters
    ----------
    spline: GustafSpline
    extrusion_vector: np.ndarray
    """
    from gustaf.spline.base import GustafSpline

    # Check input type
    if not issubclass(type(spline), GustafSpline):
        raise NotImplementedError("Extrude only works for splines")

    # Check extrusion_vector
    if extrusion_vector is not None:
        # make flat extrusion_vector
        extrusion_vector = np.asarray(extrusion_vector).ravel()
    else:
        raise ValueError("No extrusion extrusion_vector given")

    # Check extrusion_vector dimension
    # formulate correct cps
    if spline.dim == extrusion_vector.shape[0]:
        cps = spline.control_points
    elif spline.dim < extrusion_vector.shape[0]:
        expansion_dimension = extrusion_vector.shape[0] - spline.dim
        # one smaller dim is allowed
        # warn that we assume new dim is all zero
        utils.log.debug(
            f"Given extrusion vector is {expansion_dimension} dimension "
            "bigger than spline's dim. Assuming 0.0 entries for "
            "new dimension.",
        )
        cps = np.hstack(
            (
                spline.control_points,
                np.zeros((len(spline.control_points), expansion_dimension)),
            )
        )
    else:
        raise ValueError(
            "Dimension Mismatch between extrusion extrusion vector "
            "and spline."
        )

    # Start Extrusion
    spline_dict = dict()

    spline_dict["degrees"] = np.concatenate((spline.degrees, [1]))
    spline_dict["control_points"] = np.vstack((cps, cps + extrusion_vector))
    if spline.has_knot_vectors:
        spline_dict["knot_vectors"] = spline.knot_vectors + [[0, 0, 1, 1]]
    if spline.is_rational:
        spline_dict["weights"] = np.concatenate(
            (spline.weights, spline.weights)
        )

    return type(spline)(**spline_dict)


def revolved(
    spline, axis=None, center=None, angle=None, n_knot_spans=None, degree=True
):
    """Revolve spline around an axis and extend its parametric dimension.

    Parameters
    ----------
    spline : GustafSpline
      Basis-Spline to be revolved
    axis : np.ndarray
      Axis of revolution
    center : np.ndarray
      Center of revolution
    angle : float
      angle of the revolution.
    n_knot_spans : int
      number of non-zero knot-elements for result-spline (if applicable)
    degree : bool
      use degrees instead of radiant

    Returns
    -------
    spline : GustafSpline
    """

    from gustaf.spline.base import GustafSpline

    # Check input type
    if not issubclass(type(spline), GustafSpline):
        raise NotImplementedError("Revolutions only works for splines")

    # Check axis
    if axis is not None:
        # Transform into numpy array
        axis = np.asarray(axis).ravel()
        # Check Axis dimension
        if spline.control_points.shape[1] > axis.shape[0]:
            raise ValueError(
                "Dimension Mismatch between extrusion axis and spline."
            )
        elif spline.control_points.shape[1] < axis.shape[0]:
            utils.log.debug(
                "Control Point dimension is smaller than axis dimension,"
                " filling with zeros"
            )
            expansion_dimension = axis.shape[0] - spline.dim
            cps = np.hstack(
                (
                    spline.control_points,
                    np.zeros(
                        (len(spline.control_points), expansion_dimension)
                    ),
                )
            )
        else:
            cps = np.copy(spline.control_points)

        # Make sure axis is normalized

        axis_norm = np.linalg.norm(axis)
        if not np.isclose(axis_norm, 0, atol=settings.TOLERANCE):
            axis = axis / axis_norm
        else:
            raise ValueError("Axis-norm is too close to zero.")
    else:
        cps = np.copy(spline.control_points)
        if spline.control_points.shape[1] == 3:
            raise ValueError("No rotation axis given")

    # Set Problem dimension
    problem_dimension = cps.shape[1]

    # Make sure axis is ignored for 2D
    if problem_dimension == 2:
        axis = None

    # Update angle
    if angle is None:
        spline._logd("No angle given for the revolution. Using 360 degrees.")
        angle = 360

    if degree:
        angle = np.radians(angle)

    # Init center
    if center is not None:
        center = np.asarray(center).ravel()
        # Check Axis dimension
        if not (problem_dimension == center.shape[0]):
            raise ValueError(
                "Dimension Mismatch between axis and center of rotation."
            )
        cps -= center

    # The parametric dimension is independent of the revolution but the
    # rotation-matrix is only implemented for 2D and 3D problems
    if not (cps.shape[1] == 2 or cps.shape[1] == 3):
        raise NotImplementedError(
            "Sorry," "revolutions only implemented for 2D and 3D splines"
        )

    # Angle must be (0, pi) non including
    # Rotation is always performed in half steps
    PI = np.pi
    minimum_n_knot_spans = int(
        np.ceil(np.abs((angle + settings.TOLERANCE) / PI))
    )
    if (n_knot_spans) is None or (n_knot_spans < minimum_n_knot_spans):
        n_knot_spans = minimum_n_knot_spans

    if "Bezier" in spline.name:
        if n_knot_spans > 1:
            raise ValueError(
                "Revolutions are only supported for angles up to 180 "
                "degrees for Bezier type splines as they consist of only "
                "one knot span"
            )

    # Determine auxiliary values
    rot_a = angle / (2 * n_knot_spans)
    half_counter_angle = PI / 2 - rot_a
    weight = np.sin(half_counter_angle)
    factor = 1 / weight

    # Determine rotation matrix
    rotation_matrix = utils.arr.rotation_matrix_around_axis(
        axis=axis, rotation=rot_a, degree=False
    ).T

    # Start Extrusion
    spline_dict = dict()

    spline_dict["degrees"] = np.concatenate((spline.degrees, [2]))

    spline_dict["control_points"] = cps
    end_points = cps
    for i_segment in range(n_knot_spans):
        # Rotate around axis
        mid_points = np.matmul(end_points, rotation_matrix)
        end_points = np.matmul(mid_points, rotation_matrix)
        # Move away from axis using dot-product tricks
        if problem_dimension == 3:
            mp_scale = axis * np.dot(mid_points, axis).reshape(-1, 1)
            mid_points = (mid_points - mp_scale) * factor + mp_scale
        else:
            mid_points *= factor
        spline_dict["control_points"] = np.concatenate(
            (spline_dict["control_points"], mid_points, end_points)
        )

    if spline.has_knot_vectors:
        kv = [0, 0, 0]
        [kv.extend([i + 1, i + 1]) for i in range(n_knot_spans - 1)]
        spline_dict["knot_vectors"] = spline.knot_vectors + [
            kv + [n_knot_spans] * 3
        ]
    if spline.is_rational:
        mid_weights = spline.weights * weight
        spline_dict["weights"] = spline.weights
        for i_segment in range(n_knot_spans):
            spline_dict["weights"] = np.concatenate(
                (spline_dict["weights"], mid_weights, spline.weights)
            )
    else:
        utils.log.debug(
            "True revolutions are only possible for rational spline types.",
            "Creating Approximation.",
        )

    if center is not None:
        spline_dict["control_points"] += center

    return type(spline)(**spline_dict)


def from_bounds(parametric_bounds, physical_bounds):
    """Creates a minimal spline with given parametric bounds, physical bounds.
    Physical bounds can have less or equal number of
    dimension as parametric bounds. (Greater is not supported)

    Parameters
    -----------
    parametric_bounds: (2, n) array-like
    physical_bounds: (2, n) array-like

    Returns
    --------
    spline: BSpline
    """
    physical_bounds = np.asanyarray(physical_bounds).reshape(2, -1)
    parametric_bounds = np.asanyarray(parametric_bounds).reshape(2, -1)

    # get correctly sized bez box
    phys_size = physical_bounds[1] - physical_bounds[0]

    # minimal bspline
    bspline_box = box(*phys_size).bspline  # kvs are in [0, 1]
    bspline_box.cps += physical_bounds[0]  # apply offset

    # update parametric bounds
    for i, kv in enumerate(bspline_box.kvs):
        # apply scale and offset
        new_kv = (kv * parametric_bounds[1][i]) + parametric_bounds[0][i]
        bspline_box.kvs[i] = new_kv

    return bspline_box


def parametric_view(spline, axes=True):
    """Create parametric view of given spline. Previously called
    `naive_spline()`. Degrees are always 1 and knot multiplicity is not
    preserved. Returns BSpline, as BSpline and NURBS should look the same as
    parametric view.
    Will take shallow copy of underlying data of spline_data and show_options
    from original spline.

    Parameters
    -----------
    spline: BSpline or NURBS
    axes: bool
     If True, will configure axes settings, it is supported.

    Returns
    --------
    para_spline: BSpline
    """
    p_bounds = spline.parametric_bounds
    para_spline = from_bounds(
        parametric_bounds=p_bounds, physical_bounds=p_bounds
    )

    # loop through knot vectors and insert missing knots
    for i, kv in enumerate(spline.unique_knots):
        if len(kv) > 2:
            para_spline.insert_knots(i, kv[1:-1])

    # take shallow copy
    para_spline._spline_data._saved = spline.spline_data._saved.copy()
    para_spline._show_options._options[
        para_spline._show_options._backend
    ] = spline.show_options._options[spline._show_options._backend].copy()

    if axes and "axes" in spline.show_options.valid_keys():
        # configure axes
        bs = p_bounds
        bs_diff_001 = (bs[1] - bs[0]) * 0.001
        lower_b = bs[0] - bs_diff_001
        upper_b = bs[1] + bs_diff_001
        axes_config = dict(
            xtitle="u",
            ytitle="v",
            xrange=[lower_b[0], upper_b[0]],
            yrange=[lower_b[1], upper_b[1]],
            tip_size=0,
            xminor_ticks=3,
            yminor_ticks=3,
            xygrid=False,
            yzgrid=False,
        )
        if spline.para_dim == 3:
            axes_config.update(ztitle="w")
            axes_config.update(zrange=[lower_b[2], upper_b[2]])
            axes_config.update(zminor_ticks=3)
            axes_config.update(zxgrid=False)

        para_spline.show_options["axes"] = axes_config
        # it is a view, so cps won't be realistic
        para_spline.show_options["control_points"] = False
        para_spline.show_options["lighting"] = "off"

    return para_spline


def line(points):
    """Create a spline with the provided points as control points.

    Parameters
    ----------
    points: (n, d) numpy.ndarray
      npoints x ndims array of control points

    Returns
    -------
    line: BSpline
      Spline degree [1].
    """
    from gustaf import BSpline

    # lines have degree 1
    degree = 1

    cps = np.array(points)
    nknots = cps.shape[0] + degree + 1

    knots = np.concatenate(
        (
            np.full(degree, 0.0),
            np.linspace(0.0, 1.0, nknots - 2 * degree),
            np.full(degree, 1.0),
        )
    )

    spline = BSpline(
        control_points=cps, knot_vectors=[knots], degrees=[degree]
    )

    return spline


def arc(
    radius=1.0,
    angle=90.0,
    n_knot_spans=-1,
    start_angle=0.0,
    degree=True,
):
    """Creates a 1-D arc as Rational Bezier or NURBS with given radius and
    angle. The arc lies in the x-y plane and rotates around the z-axis.

    Parameters
    ----------
    radius : float, optional
      radius of the arc, defaults to 1
    angle : float, optional
      angle of the section of the arc, defaults to 90 degrees
    n_knot_spans : int
      Number of knot spans, by default minimum number for angle is used.
    start_angle : float, optional
      starting point of the angle, by default 0.
    degree: bool, optional
      degrees for angle used, by default True

    Returns
    -------
    arc: NURBS or RationalBezier
    """
    from gustaf import RationalBezier

    # Define point spline of degree 0 at starting point of the arc
    if degree:
        start_angle = np.radians(start_angle)
        angle = np.radians(angle)
    start_point = [radius * np.cos(start_angle), radius * np.sin(start_angle)]
    point_spline = RationalBezier(
        degrees=[0], control_points=[start_point], weights=[1.0]
    )
    # Bezier splines only support angles lower than 180 degrees
    if abs(angle) >= np.pi or n_knot_spans > 1:
        point_spline = point_spline.nurbs

    # Revolve - set degree to False, since all the angles are converted to rad
    arc_attrib = point_spline.create.revolved(
        angle=angle, n_knot_spans=n_knot_spans, degree=False
    ).todict()
    # Remove the first parametric dimensions, which is only a point and
    # only used for the revolution
    arc_attrib["degrees"] = list(arc_attrib["degrees"])[1:]
    if point_spline.has_knot_vectors:
        arc_attrib["knot_vectors"] = list(arc_attrib["knot_vectors"])[1:]

    return type(point_spline)(**arc_attrib)


def circle(radius=1.0, n_knot_spans=3):
    """Circle (parametric dim = 1) with radius r in the x-y plane around the
    origin. The spline has an open knot vector and degree 2.

    Parameters
    ----------
    radius : float, optional
      radius, defaults to one
    n_knots_spans : int, optional
      number of knot spans, defaults to 3

    Returns
    -------
    circle: NURBS
    """
    return arc(radius=radius, angle=360, n_knot_spans=n_knot_spans)


def box(*lengths):
    """ND box (hyper rectangle).

    Parameters
    ----------
    *lengths: list(float)

    Returns
    -------
    nd_box: Bezier
    """
    from gustaf import Bezier

    # may dim check here?
    # starting point
    nd_box = Bezier(degrees=[1], control_points=[[0], [lengths[0]]])
    # use extrude
    for i, l in enumerate(lengths[1:]):
        nd_box = nd_box.create.extruded([0] * int(i + 1) + [l])

    return nd_box


def plate(radius=1.0):
    """Creates a biquadratic 2-D spline in the shape of a plate with given
    radius.

    Parameters
    ----------
    radius : float, optional
      Radius of the plate, defaults to one

    Returns
    -------
    plate: RationalBezier
    """
    from gustaf.spline import RationalBezier

    degrees = [2, 2]
    control_points = (
        np.array(
            [
                [-0.5, -0.5],
                [0.0, -1.0],
                [0.5, -0.5],
                [-1.0, 0.0],
                [0.0, 0.0],
                [1.0, 0.0],
                [-0.5, 0.5],
                [0.0, 1.0],
                [0.5, 0.5],
            ]
        )
        * radius
    )
    weights = np.tile([1.0, 1 / np.sqrt(2)], 5)[:-1]

    return RationalBezier(
        degrees=degrees, control_points=control_points, weights=weights
    )


def disk(
    outer_radius,
    inner_radius=None,
    angle=360.0,
    n_knot_spans=4,
    degree=True,
):
    """Surface spline describing a potentially hollow disk with quadratic
    degree along curved dimension and linear along thickness. The angle
    describes the returned part of the disk.

    Parameters
    ----------
    outer_radius : float
      Outer radius of the disk
    inner_radius : float, optional
      Inner radius of the disk, in case of hollow disk, by default 0.
    angle : float, optional
      Rotational angle, by default 360. describing a complete revolution
    n_knot_spans : int, optional
      Number of knot spans, by default 4

    Returns
    -------
    disk: NURBS
      Surface NURBS of degrees (1,2)
    """

    from gustaf.spline import NURBS

    if inner_radius is None:
        inner_radius = 0.0

    cps = np.array([[inner_radius, 0.0], [outer_radius, 0.0]])
    weights = np.ones([cps.shape[0]])
    knots = np.repeat([0.0, 1.0], 2)

    return NURBS(
        control_points=cps, knot_vectors=[knots], degrees=[1], weights=weights
    ).create.revolved(angle=angle, n_knot_spans=n_knot_spans, degree=degree)


def torus(
    torus_radius,
    section_outer_radius,
    section_inner_radius=None,
    torus_angle=None,
    section_angle=None,
    section_n_knot_spans=4,
    torus_n_knot_spans=4,
    degree=True,
):
    """Creates a volumetric NURBS spline describing a torus revolved around the
    x-axis. Possible cross-sections are plate, disk (yielding a tube) and
    section of a disk.

    Parameters
    ----------
    torus_radius : float
      Radius of the torus
    section_outer_radius : float
      Radius of the section of the torus
    section_inner_radius : float, optional
      Inner radius in case of hollow torus, by default 0.
    torus_angle : float, optional
      Rotational angle of the torus, by default None, giving a complete
      revolution
    section_angle : float, optional
      Rotational angle, by default None, yielding a complete revolution
    section_n_knot_spans : float, optional
      Number of knot spans along the cross-section, by default 4
    torus_n_knot_spans : float, optional
      Number of knot spans along the torus, by default 4

    Returns
    -------
    torus: NURBS
      Volumetric spline in the shape of a torus with degrees (1,2,2)
    """

    if torus_angle is None:
        torus_angle = 2 * np.pi
        degree = False

    if section_angle is None:
        section_angle = 2 * np.pi
        section_angle_flag = False
        degree = False
    else:
        section_angle_flag = True

    if section_inner_radius is None:
        section_inner_radius = 0
        section_inner_radius_flag = False
    else:
        section_inner_radius_flag = True

    # Create the cross-section
    if not section_angle_flag and not section_inner_radius_flag:
        cross_section = plate(section_outer_radius)
        # For more than 180 degree only NURBS can be used
        if abs(torus_angle) >= np.pi:
            cross_section = cross_section.nurbs
    else:
        cross_section = disk(
            outer_radius=section_outer_radius,
            inner_radius=section_inner_radius,
            n_knot_spans=section_n_knot_spans,
            angle=section_angle,
            degree=degree,
        )

    # Create a surface spline representing a disk and move it from the origin
    cross_section.control_points[:, 1] += torus_radius

    return cross_section.create.revolved(
        axis=[1.0, 0, 0],
        center=np.zeros(3),
        angle=torus_angle,
        n_knot_spans=torus_n_knot_spans,
        degree=degree,
    )


def sphere(
    outer_radius,
    inner_radius=None,
    angle=360.0,
    n_knot_spans=-1,
    degree=True,
):
    """Creates a volumetric spline describing a sphere with radius R.

    Parameters
    ----------
    outer_radius : float
      Outer radius of the sphere
    inner_radius : float, optional
      Inner radius of the potentially hollow sphere.
    angle : float
      Rotational angle around x-axis, by default each 360
      (describing a complete revolution)
    n_knot_spans : int
      Number of knot spans

    Returns
    -------
    sphere: NURBS
      Volumetric NURBS with degrees (1,2,2)
    """

    if inner_radius is None:
        sphere = plate(outer_radius).nurbs.create.revolved(
            axis=[1, 0, 0],
            center=[0, 0, 0],
            angle=angle,
            n_knot_spans=n_knot_spans,
            degree=degree,
        )
    else:
        inner_radius = float(inner_radius)
        sphere = disk(outer_radius, inner_radius).nurbs.create.revolved(
            angle=angle, n_knot_spans=n_knot_spans, degree=degree
        )
    return sphere


def cone(
    outer_radius,
    height,
    inner_radius=None,
    volumetric=True,
    angle=360.0,
    degree=True,
):
    """Creates a cone with circular base.

    Parameters
    ----------
    radius : float
      Radius of the base
    height : float
      Height of the cone
    volumetric : bool, optional
      Parameter whether surface or volume spline, by default True
    angle : float
      Rotation angle in degrees, only used for solid model

    Returns
    -------
    cone: NURBS
      Volumetric or surface NURBS describing a cone
    """

    if volumetric:
        ground = disk(
            outer_radius, inner_radius=inner_radius, angle=angle, degree=degree
        )
    else:
        ground = circle(outer_radius)

    # Extrude in z
    cone = ground.create.extruded([0, 0, height])
    # Move all upper control points to one
    cone.control_points[np.isclose(cone.control_points[:, -1], height)] = [
        0,
        0,
        height,
    ]

    return cone


def pyramid(width, length, height):
    """Creates a volumetric spline in the shape of a pyramid with linear degree
    in every direction.

    Parameters
    ----------
    width : float
      Dimension of base in x-axis
    length : float
      Dimension of base in y-axis
    height : float
      Height in z-direction

    Returns
    -------
    pyramid: BSpline
      Volumetric linear spline in the shape of a pyramid
    """

    # Create box
    p = box(width, length, height)

    # Collapse all upper points on one control point
    p.control_points[4:, :] = [width / 2, length / 2, height]

    return p


class Creator:
    """Helper class to build new splines from existing geometries.

    Examples
    ---------
    >>> my_spline = <your-spline>
    >>> spline_faces = my_spline.create.extrude(vector=[3,1,3])
    """

    def __init__(self, spl):
        self.spline = spl

    def extruded(self, *args, **kwargs):
        return extruded(self.spline, *args, **kwargs)

    def revolved(self, *args, **kwargs):
        return revolved(self.spline, *args, **kwargs)

    def parametric_view(self, *args, **kwargs):
        return parametric_view(self.spline, *args, **kwargs)



"""gustaf/create/faces.py
Routines to create faces.
"""

import numpy as np

from gustaf import create, settings, utils
from gustaf.faces import Faces


def box(
    bounds=[[0, 0], [1, 1]], resolutions=[2, 2], simplex=False, backslash=False
):
    """Create structured quadrangle or triangle block mesh.

    Parameters
    -----------
    bounds: (2, 2) array
        Minimum and maximum coordinates.
    resolutions: (2) array
        Vertex count in each dimension.
    simplex: boolean
        If true, Mesh will be triangular (simplex).

    Returns
    --------
    face_mesh: Volumes
    """

    if np.array(bounds).shape != (2, 2):
        raise ValueError("Bounds must have a dimension of (2, 2).")
    if len(resolutions) != 2:
        raise ValueError("Resolutions must have two entries.")
    if not np.greater(resolutions, 1).all():
        raise ValueError("All resolution values must be at least 2.")

    if simplex:
        # create quad mesh as basis
        quad_mesh = box(bounds=bounds, resolutions=resolutions)

        # turn into triangles
        face_mesh = to_simplex(quad_mesh, backslash)

    else:
        vertex_mesh = create.vertices.raster(bounds, resolutions)
        connectivity = utils.connec.make_quad_faces(resolutions)
        face_mesh = Faces(vertex_mesh.vertices, connectivity)

    return face_mesh


def to_simplex(quad, backslash=False):
    """Given quad faces, diagonalize them to turn them into triangles.

    If quad is counterclockwise (CCW), triangle will also be CCW and
    vice versa. Will return a tri-mesh, if input is triangular.
    Default diagonalization looks like this:

    .. code-block::

        (3) *---* (2)
            |  /|
            | / |
            |/  |
        (0) *---* (1)

    resembling 'slash'.

    .. code-block::

        (3) *---* (2)
            |\\  |
            | \\ |
            |  \\|
        (0) *---* (1)

    resembling 'backslash'.

    If you want to diagonalize the other way, set `backslash=True`.

    Parameters
    ----------
    quad: Faces
        Faces representation which is to be converted from a cubic mesh into a
        simplex mesh.
    backslash: bool

    Returns
    --------
    tri: Faces
        Simplexifyed mesh.
    """

    if not isinstance(quad, Faces):
        raise ValueError(
            "Input to to_simplex needs to be of type Faces, but it's "
            + type(quad)
        )

    if quad.whatami.startswith("quad"):
        # split variants
        split_slash = [[0, 1, 2], [2, 3, 0]]
        split_backslash = [[0, 1, 3], [3, 1, 2]]

        quad_faces = quad.faces
        tf_half = int(quad_faces.shape[0])
        tri_faces = np.full((tf_half * 2, 3), -1, dtype=settings.INT_DTYPE)

        split = split_backslash if backslash else split_slash

        tri_faces[:tf_half] = quad_faces[:, split[0]]
        tri_faces[tf_half:] = quad_faces[:, split[1]]

        tri = Faces(
            vertices=quad.vertices.copy(),
            faces=tri_faces,
        )
    else:
        tri = quad
        utils.log.debug(
            "Non quadrilateral mesh provided, return original" " mesh."
        )

    return tri



"""Import a meshio based mesh,

Export can only happen after it is possible to save and define boundaries in
gustaf.
"""
import pathlib
from typing import Union

import numpy as np

from gustaf._typing import MESH_TYPES
from gustaf.faces import Faces
from gustaf.helpers.raise_if import ModuleImportRaiser
from gustaf.volumes import Volumes

try:
    import meshio
except ModuleNotFoundError as err:
    meshio = ModuleImportRaiser("meshio", err)
    # from meshio import Mesh as MeshioMesh


def load(fname: Union[str, pathlib.Path]) -> MESH_TYPES:
    """Load mesh in meshio format. Loads vertices and their connectivity.
    Currently cannot process boundary.

    Note
    -----
    This is more or less a direct copy from the original gustav implementation.
    A lot of the meshio information are lost. When boundaries and multi-patch
    definitions are added this needs to be revisited and extended.

    Parameters
    ------------
    fname: str | pathlib.Path

    Returns
    --------
    MESH_TYPES
    """
    mesh_type = Faces

    fname = pathlib.Path(fname)
    if not (fname.exists() and fname.is_file()):
        raise ValueError(
            "The given file does not point to file. The given path is: "
            f"{fname.resolve()}"
        )

    meshio_mesh: meshio.Mesh = meshio.read(fname)
    vertices = meshio_mesh.points

    # check for 2D mesh
    # Try for triangle grid
    cells = meshio_mesh.get_cells_type("triangle")
    # If  no triangle elements, try for square
    if len(cells) == 0:
        cells = meshio_mesh.get_cells_type("quad")
    if not len(cells) > 0:
        # 3D mesh
        mesh_type = Volumes
        cells = meshio_mesh.get_cells_type("tetra")
        if len(cells) == 0:
            cells = meshio_mesh.get_cells_type("hexahedron")

    for i, cell in enumerate(cells):
        if i == 0:
            elements = cell.data
        else:
            elements = np.vstack((elements, cell.data))

    mesh = mesh_type(vertices=vertices, elements=elements)

    return mesh


def export(mesh: MESH_TYPES, fname: Union[str, pathlib.Path]):
    """Currently not implemented function.

    Parameters
    ------------
    mesh: MESH_TYPES
      Mesh to be exported.
    fname: Union[str, pathlib.Path]
      File to save the mesh in.

    Raises
    -------
    NotImplementedError: This method is currently not implemented.
    """
    raise NotImplementedError



import pathlib
from typing import Union

from gustaf._typing import MESH_TYPES
from gustaf.io import meshio, mfem, mixd


def load(fname: Union[str, pathlib.Path]) -> MESH_TYPES:
    """Load function for all supported file formats.

    This function tries to guess the correct io module for the given file.

    Parameters
    -----------
    fname: Union[str, pathlib.Path]
        Name of the file to be loaded.

    Returns
    --------
    MESH_TYPES:
        Loaded mesh.
    """
    extensions_to_load_functions = {
        ".mixd": mixd.load,
        ".mfem": mfem.load,
        ".msh": meshio.load,
    }

    fname = pathlib.Path(fname).resolve()

    if fname.suffix in extensions_to_load_functions:
        return extensions_to_load_functions[fname.suffix](fname)

    else:
        raise ValueError(
            f"Failed to load given file with '{fname.suffix}' extension. "
            "Valid extensions are: "
            f"{tuple(extensions_to_load_functions.keys())}."
        )



"""gustaf/spline/base.py.

Base for splines. Contains show and inherited classes from `spline`.
"""
from copy import deepcopy

import numpy as np
import splinepy

from gustaf import show as showmodule
from gustaf import utils
from gustaf._base import GustafBase
from gustaf.helpers.data import SplineData
from gustaf.spline import visualize
from gustaf.spline.create import Creator
from gustaf.spline.extract import Extractor
from gustaf.spline.proximity import Proximity


def show(spline, **kwargs):
    """Shows splines with various options.

    They are excessively listed, so that it can be adjustable.

    Parameters
    -----------
    spline: BSpline or NURBS
    resolutions: int or (spline.para_dim,) array-like
    control_points: bool
    knots: bool
    fitting_queries: bool
    return_discrete: bool
      Return dict of gustaf discrete objects, for example,
      {Vertices, Edges, Faces}, instead of opening a window
    return_showable: bool
      Return dict of showable objects.
    parametric_space: bool
      Only relevant for `vedo` backend.
    c: str
      Default is None. Black for curves, else green.
    alpha: float
    lighting: str
    control_point_ids: bool
    color_spline: str
    cmap: str

    Returns
    --------
    things_to_show: dict
      iff return_discrete==True, dict of gustaf objects that are showable.
      iff return_showable==True, dict of backend objects that are showable.
    """
    # Showing is only possible for following splines
    allowed_dim_combo = (
        (1, 2),
        (1, 3),
        (2, 2),
        (2, 3),
        (3, 3),
    )
    if (spline.para_dim, spline.dim) not in allowed_dim_combo:
        raise ValueError("Sorry, can't show given spline.")

    if kwargs:
        orig_show_options = spline.show_options
        spline._show_options = spline.__show_option__(spline)
        orig_show_options.copy_valid_options(spline.show_options)
        for key, value in kwargs.items():
            try:
                spline.show_options[key] = value
            except BaseException:
                utils.log.debug(
                    f"Skipping invalid option {key} for "
                    f"{spline.show_options._helps}."
                )
                continue

    # Prepare things to show dict.
    things_to_show = visualize.make_showable(spline)

    para_space = kwargs.pop("parametric_space", False)
    if para_space:
        p_spline = spline.create.parametric_view()
        things_to_show["parametric_spline"] = p_spline

    if kwargs.get("return_discrete", False):
        return things_to_show

    if kwargs.get("return_showable", False):
        return {key: value.showable() for key, value in things_to_show.items()}

    if para_space:
        things_to_show.pop("parametric_spline")

        return showmodule.show_vedo(
            ["Parametric Space", p_spline],
            ["Physical Space", *things_to_show.values()],
            **kwargs,
        )

    return showmodule.show_vedo(things_to_show, **kwargs)


class GustafSpline(GustafBase):
    __show_option__ = visualize.SplineShowOption

    def __init__(self):
        """Constructor as abstractmethod.

        This needs to be inherited first to make sure duplicating
        functions properly override splinepy.Spline
        """
        self._extractor = Extractor(self)
        self._proximity = Proximity(self)
        self._creator = Creator(self)
        self._show_options = self.__show_option__(self)
        self._spline_data = SplineData(self)

    @property
    def show_options(self):
        """
        Show option manager for splines.

        Parameters
        ----------
        None

        Returns
        -------
        show_options: SplineShowOption
        """
        return self._show_options

    @property
    def spline_data(self):
        """
        Spline data helper for splines.

        Parameters
        ----------
        None

        Returns
        -------
        spline_data: SplineData
        """
        return self._spline_data

    @property
    def extract(self):
        """Returns spline extractor. Can directly perform extractions available
        at `gustaf/spline/extract.py`. For more info, take a look at
        `gustaf/spline/extract.py`: Extractor.

        Examples
        ---------
        >>> spline_faces = spline.extract.faces()

        Parameters
        -----------
        None

        Returns
        --------
        spline_extractor: Extractor
        """
        return self._extractor

    @property
    def create(self):
        """Returns spline creator Can be used to create new splines using
        geometric relations.

        Examples
        --------
        >>> prism = spline.create.extrude(axis=[1,4,1])

        Parameters
        ----------
        None

        Returns
        spline.Creator
        """
        return self._creator

    @property
    def proximity(self):
        """Returns spline proximity helper. Can directly perform proximity
        queries available at `gustaf/spline/proximity.py`. For more info, take
        a look at `gustaf/spline/proximity.py`: Proximity.

        Examples
        ---------
        >>> closest_cp_ids = spline.proximity.closest_control_points(queries)
        >>> closest_cp_ids, distances =\
        ...    spline.proximity.closest_control_points(
        ...       queries,
        ...       return_distances=True
        ...    )

        Parameters
        -----------
        None

        Returns
        --------
        spline_proximity: Proximity
        """
        return self._proximity

    def show(self, **kwargs):
        """Equivalent to `gustaf.spline.base.show(**kwargs)`"""
        return show(self, **kwargs)

    def showable(self, **kwargs):
        """Equivalent to
        `gustaf.spline.base.show(return_showable=True,**kwargs)`"""
        return show(self, return_showable=True, **kwargs)

    def copy(self):
        """tmp copy from splinepy until #89 merges"""
        new = type(self)()
        new.new_core(**self._data["properties"], properties_round_trip=False)
        new._data = deepcopy(self._data)
        return new


class Bezier(GustafSpline, splinepy.Bezier):
    def __init__(
        self,
        degrees=None,
        control_points=None,
        spline=None,
    ):
        """Bezier of gustaf. Inherited from splinepy.Bezier and GustafSpline.

        Attributes
        -----------
        extract: Extractor
        create: Creator
        proximity: Proximity

        Parameters
        -----------
        degrees: (para_dim,) list-like
        control_points: (m, dim) list-like

        Returns
        --------
        None
        """
        splinepy.Bezier.__init__(
            self, degrees=degrees, control_points=control_points, spline=spline
        )
        GustafSpline.__init__(self)

    @property
    def bezier(self):
        """Returns same parametric representation as Bezier Spline.

        Parameters
        ----------
        None

        Returns
        -------
        same : Bezier
        """
        return self.copy()

    @property
    def rationalbezier(self):
        """Returns same parametric representation as Rational Bezier Spline.

        Parameters
        ----------
        None

        Returns
        -------
        same : RationalBezier
        """
        return RationalBezier(
            degrees=self.degrees,
            control_points=self.control_points,
            weights=np.ones(self.control_points.shape[0]),
        )

    @property
    def bspline(self):
        """Returns same parametric representation as BSpline.

        Parameters
        -----------
        None

        Returns
        --------
        same_bspline : BSpline
        """
        return BSpline(
            degrees=self.degrees,
            control_points=self.control_points,
            knot_vectors=[
                [0] * (self.degrees[i] + 1) + [1] * (self.degrees[i] + 1)
                for i in range(self.para_dim)
            ],
        )

    @property
    def nurbs(self):
        """Returns same parametric representation as nurbs.

        Parameters
        -----------
        None

        Returns
        --------
        same_nurbs: NURBS
        """
        return self.bspline.nurbs


class RationalBezier(GustafSpline, splinepy.RationalBezier):
    def __init__(
        self,
        degrees=None,
        control_points=None,
        weights=None,
        spline=None,
    ):
        """Rational Bezier of gustaf. Inherited from splinepy.RationalBezier
        and GustafSpline.

        Attributes
        -----------
        extract: Extractor
        create: Creator
        proximity: Proximity

        Parameters
        -----------
        degrees: (para_dim,) list-like
        control_points: (m, dim) list-like
        weights : (m) list-like

        Returns
        --------
        None
        """
        splinepy.RationalBezier.__init__(
            self,
            degrees=degrees,
            control_points=control_points,
            weights=weights,
            spline=spline,
        )
        GustafSpline.__init__(self)

    @property
    def rationalbezier(self):
        """Returns same parametric representation as Rational Bezier Spline.

        Parameters
        ----------
        None

        Returns
        -------
        same : RationalBezier
        """
        return self.copy()

    @property
    def nurbs(self):
        """Returns same parametric representation as nurbs.

        Parameters
        -----------
        None

        Returns
        --------
        same_nurbs: NURBS
        """
        return NURBS(
            degrees=self.degrees,
            control_points=self.control_points,
            knot_vectors=[
                [0] * (self.degrees[i] + 1) + [1] * (self.degrees[i] + 1)
                for i in range(self.para_dim)
            ],
            weights=self.weights,
        )


class BSpline(GustafSpline, splinepy.BSpline):
    def __init__(
        self,
        degrees=None,
        knot_vectors=None,
        control_points=None,
        spline=None,
    ):
        """BSpline of gustaf. Inherited from splinepy.BSpline and GustafSpline.

        Attributes
        -----------
        extract: Extractor
        create: Creator
        proximity: Proximity

        Parameters
        -----------
        degrees: (para_dim,) list-like
        knot_vectors: (para_dim, ...) list
        control_points: (m, dim) list-like

        Returns
        --------
        None
        """
        splinepy.BSpline.__init__(
            self,
            degrees=degrees,
            knot_vectors=knot_vectors,
            control_points=control_points,
            spline=spline,
        )

        GustafSpline.__init__(self)

    @property
    def bspline(self):
        """Returns same parametric representation as BSpline.

        Parameters
        -----------
        None

        Returns
        --------
        same_bspline : BSpline
        """
        return self.copy()

    @property
    def nurbs(self):
        """Returns same nurbs. Overwrites one from splinepy to return correct
        type.

        Parameters
        -----------
        None

        Returns
        --------
        same_nurbs: NURBS
        """
        from copy import deepcopy

        return NURBS(
            degrees=deepcopy(self.degrees),
            knot_vectors=deepcopy(self.knot_vectors),
            control_points=deepcopy(self.control_points),
            # fix dtype to match splinepy's dtype.
            weights=np.ones(self.control_points.shape[0], dtype="float64"),
        )


class NURBS(GustafSpline, splinepy.NURBS):
    def __init__(
        self,
        degrees=None,
        knot_vectors=None,
        control_points=None,
        weights=None,
        spline=None,
    ):
        """NURBS of gustaf. Inherited from splinepy.NURBS.

        Attributes
        -----------
        extract: Extractor
        create: Creator
        proximity: Proximity

        Parameters
        -----------
        degrees: (para_dim,) list-like
        knot_vectors: (para_dim,) list
        control_points: (m, dim) list-like
        weights: (m, 1) list-like

        Returns
        --------
        None
        """
        splinepy.NURBS.__init__(
            self,
            degrees=degrees,
            knot_vectors=knot_vectors,
            control_points=control_points,
            weights=weights,
            spline=spline,
        )
        GustafSpline.__init__(self)

    @property
    def _mfem_ids(self):
        """Returns mfem index mapping. For ease of use.

        Parameters
        -----------
        None

        Returns
        --------
        None
        """
        if self.para_dim != 2:
            raise NotImplementedError(
                "Sorry, only available for para_dim = 2 splines"
            )

        gustaf2mfem, mfem2gustaf = splinepy.io.mfem.mfem_index_mapping(
            self.para_dim,
            self.degrees,
            self.knot_vectors,
        )

        return gustaf2mfem, mfem2gustaf

    @property
    def nurbs(self):
        """Returns same parametric representation as nurbs.

        Parameters
        -----------
        None

        Returns
        --------
        same_nurbs: NURBS
        """
        return self.copy()


def from_mfem(nurbs_dict):
    """Construct a gustaf NURBS. Reorganizes control points and weights.

    Parameters
    -----------
    nurbs_dict: dict

    Returns
    --------
    nurbs: NURBS
    """
    _, m2gus = splinepy.io.mfem.mfem_index_mapping(
        len(nurbs_dict["degrees"]),
        nurbs_dict["degrees"],
        nurbs_dict["knot_vectors"],
    )

    return NURBS(
        degrees=nurbs_dict["degrees"],
        knot_vectors=nurbs_dict["knot_vectors"],
        control_points=nurbs_dict["control_points"][m2gus],
        weights=nurbs_dict["weights"][m2gus],
    )


def load_splines(fname):
    """Loads and creates gustaf NURBS. Does not perform any check or tests.

    Parameters
    -----------
    fname: str

    Returns
    --------
    gussplines: list
    """
    # first get dict_splines using splinepy
    dictsplines = splinepy.load_splines(fname, as_dict=True)

    # try to initialize with correct spline type
    gussplines = list()
    for dics in dictsplines:
        is_bspline = "knot_vectors" in dics
        is_nurbs = "weights" in dics

        if is_nurbs:
            gussplines.append(NURBS(**dics))
        elif is_bspline:
            gussplines.append(BSpline(**dics))
        else:
            gussplines.append(Bezier(**dics))

    return gussplines



"""gustaf/spline/__init__.py.

Interface for `splinepy` library and additional operations. Everything
related to `splinepy` is here, which includes its load function.
However, since create.spline does not rely on `splinepy`, it is not
here.
"""

import splinepy
from splinepy import io

from gustaf.helpers.data import SplineDataAdaptor
from gustaf.spline import base, create, extract, ffd, microstructure, visualize
from gustaf.spline.base import (
    NURBS,
    Bezier,
    BSpline,
    RationalBezier,
    from_mfem,
    load_splines,
    show,
)

# overwrite name to type map
splinepy.settings.NAME_TO_TYPE = dict(
    Bezier=Bezier,
    RationalBezier=RationalBezier,
    BSpline=BSpline,
    NURBS=NURBS,
)

# a shortcut
NAME_TO_TYPE = splinepy.settings.NAME_TO_TYPE

__all__ = [
    "base",
    "create",
    "extract",
    "Bezier",
    "RationalBezier",
    "BSpline",
    "NURBS",
    "SplineDataAdaptor",
    "show",
    "from_mfem",
    "load_splines",
    "ffd",
    "microstructure",
    "io",
    "visualize",
]



"""gustaf/create/vertices.py.

Routines to create vertices.
"""

import numpy as np

from gustaf.vertices import Vertices


def raster(
    bounds,
    resolutions,
):
    """Simple wrapper of np.mgrid to extract raster points of desired bounds
    and resolutions.

    Parameters
    -----------
    bounds: (2, d) array-like
      float
    resolutions: (d,) array-like
      int. It will be casted to int.

    Returns
    --------
    raster_vertices: Vertices
    """
    if len(resolutions) != len(bounds[0]) == len(bounds[1]):
        raise ValueError("Length of resolutions and bounds should match.")

    slices = list()
    for b0, b1, r in zip(bounds[0], bounds[1], resolutions):
        slices.append(slice(b0, b1, r * 1j))

    # Organize it nicely: 2D np.ndarray with shape (prod(resolutions), dim)
    points = np.mgrid[slices].T.reshape(-1, len(resolutions))

    return Vertices(points)



import numpy as np

from gustaf.spline import base
from gustaf.spline.microstructure.tiles.tilebase import TileBase


class CrossTile2D(TileBase):
    def __init__(self):
        """Simple crosstile with linear-quadratic branches and a trilinear
        center spline."""
        self._dim = 2
        self._evaluation_points = np.array(
            [
                [0.0, 0.5],
                [1.0, 0.5],
                [0.5, 0.0],
                [0.5, 1.0],
            ]
        )
        self._parameter_space_dimension = 1

    def closing_tile(
        self,
        parameters=None,
        closure=None,
        boundary_width=0.1,
        filling_height=0.5,
        **kwargs,
    ):
        """Create a closing tile to match with closed surface.

        Parameters
        ----------
        parameters : tuple(np.ndarray)
          radii of fitting cylinder at evaluation points
        closure : int
          parametric dimension that needs to be closed. Positiv values mean
          that minimum parametric dimension is requested. That means,
          i.e. -2 closes the tile at maximum z-coordinate.
          (must currently be either -2 or 2)
        boundary_width : float
          with of the boundary surronding branch
        filling_height : float
          portion of the height that is filled in parametric domain

        Returns
        -------
        list_of_splines : list
        """
        # Check parameters
        if closure is None:
            raise ValueError("No closing direction given")

        if parameters is None:
            self._logd("Tile request is not parametrized, setting default 0.2")
            parameters = tuple([np.ones(6) * 0.2])
        parameters = parameters[0]
        if not (np.all(parameters > 0) and np.all(parameters < 0.5)):
            raise ValueError("Thickness out of range (0, .5)")

        if not (0.0 < float(boundary_width) < 0.5):
            raise ValueError("Boundary Width is out of range")

        if not (0.0 < float(filling_height) < 1.0):
            raise ValueError("Filling must  be in (0,1)")

        # Constant auxiliary values
        inv_boundary_width = 1.0 - boundary_width
        inv_filling_height = 1.0 - filling_height
        ctps_mid_height_top = (1 + filling_height) * 0.5
        ctps_mid_height_bottom = 1.0 - ctps_mid_height_top
        v_one_half = 0.5
        v_one = 1.0
        v_zero = 0.0

        spline_list = []
        if closure == "x_min":
            # Minimum x position
            branch_thickness = parameters[1]

            block0_ctps = np.array(
                [
                    [v_zero, v_zero],
                    [filling_height, v_zero],
                    [v_zero, boundary_width],
                    [filling_height, boundary_width],
                ]
            )

            block1_ctps = np.array(
                [
                    [v_zero, boundary_width],
                    [filling_height, boundary_width],
                    [v_zero, inv_boundary_width],
                    [filling_height, inv_boundary_width],
                ]
            )

            block2_ctps = np.array(
                [
                    [v_zero, inv_boundary_width],
                    [filling_height, inv_boundary_width],
                    [v_zero, v_one],
                    [filling_height, v_one],
                ]
            )

            branch_ctps = np.array(
                [
                    [filling_height, boundary_width],
                    [ctps_mid_height_top, v_one_half - branch_thickness],
                    [v_one, v_one_half - branch_thickness],
                    [filling_height, inv_boundary_width],
                    [ctps_mid_height_top, v_one_half + branch_thickness],
                    [v_one, v_one_half + branch_thickness],
                ]
            )

            spline_list.append(
                base.Bezier(degrees=[1, 1], control_points=block0_ctps)
            )

            spline_list.append(
                base.Bezier(degrees=[1, 1], control_points=block1_ctps)
            )

            spline_list.append(
                base.Bezier(degrees=[1, 1], control_points=block2_ctps)
            )

            spline_list.append(
                base.Bezier(degrees=[2, 1], control_points=branch_ctps)
            )
            return spline_list
        elif closure == "x_max":
            # Maximum x position
            branch_thickness = parameters[0]

            block0_ctps = np.array(
                [
                    [inv_filling_height, v_zero],
                    [v_one, v_zero],
                    [inv_filling_height, boundary_width],
                    [v_one, boundary_width],
                ]
            )

            block1_ctps = np.array(
                [
                    [inv_filling_height, boundary_width],
                    [v_one, boundary_width],
                    [inv_filling_height, inv_boundary_width],
                    [v_one, inv_boundary_width],
                ]
            )

            block2_ctps = np.array(
                [
                    [inv_filling_height, inv_boundary_width],
                    [v_one, inv_boundary_width],
                    [inv_filling_height, v_one],
                    [v_one, v_one],
                ]
            )

            branch_ctps = np.array(
                [
                    [0, v_one_half - branch_thickness],
                    [ctps_mid_height_bottom, v_one_half - branch_thickness],
                    [inv_filling_height, boundary_width],
                    [v_zero, v_one_half + branch_thickness],
                    [ctps_mid_height_bottom, v_one_half + branch_thickness],
                    [inv_filling_height, inv_boundary_width],
                ]
            )

            spline_list.append(
                base.Bezier(degrees=[1, 1], control_points=block0_ctps)
            )

            spline_list.append(
                base.Bezier(degrees=[1, 1], control_points=block1_ctps)
            )

            spline_list.append(
                base.Bezier(degrees=[1, 1], control_points=block2_ctps)
            )

            spline_list.append(
                base.Bezier(degrees=[2, 1], control_points=branch_ctps)
            )
            return spline_list
        elif closure == "y_min":
            # Minimum y position
            branch_thickness = parameters[3]

            block0_ctps = np.array(
                [
                    [v_zero, v_zero],
                    [boundary_width, v_zero],
                    [v_zero, filling_height],
                    [boundary_width, filling_height],
                ]
            )

            block1_ctps = np.array(
                [
                    [boundary_width, v_zero],
                    [inv_boundary_width, v_zero],
                    [boundary_width, filling_height],
                    [inv_boundary_width, filling_height],
                ]
            )

            block2_ctps = np.array(
                [
                    [inv_boundary_width, v_zero],
                    [v_one, v_zero],
                    [inv_boundary_width, filling_height],
                    [v_one, filling_height],
                ]
            )

            branch_ctps = np.array(
                [
                    [boundary_width, filling_height],
                    [inv_boundary_width, filling_height],
                    [v_one_half - branch_thickness, ctps_mid_height_top],
                    [v_one_half + branch_thickness, ctps_mid_height_top],
                    [v_one_half - branch_thickness, v_one],
                    [v_one_half + branch_thickness, v_one],
                ]
            )

            spline_list.append(
                base.Bezier(degrees=[1, 1], control_points=block0_ctps)
            )

            spline_list.append(
                base.Bezier(degrees=[1, 1], control_points=block1_ctps)
            )

            spline_list.append(
                base.Bezier(degrees=[1, 1], control_points=block2_ctps)
            )

            spline_list.append(
                base.Bezier(degrees=[1, 2], control_points=branch_ctps)
            )
            return spline_list
        elif closure == "y_max":
            # Maximum y position
            branch_thickness = parameters[2]

            block0_ctps = np.array(
                [
                    [v_zero, inv_filling_height],
                    [boundary_width, inv_filling_height],
                    [v_zero, v_one],
                    [boundary_width, v_one],
                ]
            )

            block1_ctps = np.array(
                [
                    [boundary_width, inv_filling_height],
                    [inv_boundary_width, inv_filling_height],
                    [boundary_width, v_one],
                    [inv_boundary_width, v_one],
                ]
            )

            block2_ctps = np.array(
                [
                    [inv_boundary_width, inv_filling_height],
                    [v_one, inv_filling_height],
                    [inv_boundary_width, v_one],
                    [v_one, v_one],
                ]
            )

            branch_ctps = np.array(
                [
                    [v_one_half - branch_thickness, v_zero],
                    [v_one_half + branch_thickness, v_zero],
                    [v_one_half - branch_thickness, ctps_mid_height_bottom],
                    [v_one_half + branch_thickness, ctps_mid_height_bottom],
                    [boundary_width, inv_filling_height],
                    [inv_boundary_width, inv_filling_height],
                ]
            )

            spline_list.append(
                base.Bezier(degrees=[1, 1], control_points=block0_ctps)
            )

            spline_list.append(
                base.Bezier(degrees=[1, 1], control_points=block1_ctps)
            )

            spline_list.append(
                base.Bezier(degrees=[1, 1], control_points=block2_ctps)
            )

            spline_list.append(
                base.Bezier(degrees=[1, 2], control_points=branch_ctps)
            )
            return spline_list
        else:
            raise NotImplementedError(
                "Requested closing dimension is not supported"
            )

    def create_tile(self, parameters=None, center_expansion=1.0, **kwargs):
        """Create a microtile based on the parameters that describe the branch
        thicknesses.

        Thickness parameters are used to describe the inner radius of the
        outward facing branches

        Parameters
        ----------
        parameters : tuple(np.array)
            only first entry is used, defines the internal radii of the
            branches
        center_expansion : float
            thickness of center is expanded by a factor
        Returns
        -------
        microtile_list : list(splines)
        """

        if not isinstance(center_expansion, float):
            raise ValueError("Invalid Type")
        if not ((center_expansion > 0.5) and (center_expansion < 1.5)):
            raise ValueError("Center Expansion must be in (.5,1.5)")
        max_radius = min(0.5, (0.5 / center_expansion))
        # set to default if nothing is given
        if parameters is None:
            self._logd("Setting branch thickness to default 0.2")
            parameters = tuple([np.ones(4) * 0.2])
        [x_min_r, x_max_r, y_min_r, y_max_r] = parameters[0].tolist()
        for radius in [x_min_r, x_max_r, y_min_r, y_max_r]:
            if not isinstance(radius, float):
                raise ValueError("Invalid type")
            if not (radius > 0 and radius < max_radius):
                raise ValueError(
                    f"Radii must be in (0,{max_radius}) for "
                    f"center_expansion {center_expansion}"
                )

        # center radius
        center_r = (
            (x_min_r + x_max_r + y_min_r + y_max_r) / 4.0 * center_expansion
        )
        hd_center = 0.5 * (0.5 + center_r)
        one_half = 0.5

        # Init return value
        spline_list = []

        # Create the center-tile
        center_points = np.array(
            [
                [-center_r, -center_r],
                [center_r, -center_r],
                [-center_r, center_r],
                [center_r, center_r],
            ]
        ) + np.array([one_half, one_half])

        y_min_ctps = np.array(
            [
                [-y_min_r, -one_half],
                [y_min_r, -one_half],
                [-y_min_r, -hd_center],
                [y_min_r, -hd_center],
                [-center_r, -center_r],
                [center_r, -center_r],
            ]
        ) + np.array([one_half, one_half])

        y_max_ctps = np.array(
            [
                [-center_r, center_r],
                [center_r, center_r],
                [-y_max_r, hd_center],
                [y_max_r, hd_center],
                [-y_max_r, one_half],
                [y_max_r, one_half],
            ]
        ) + np.array([one_half, one_half])

        x_min_ctps = np.array(
            [
                [-one_half, -x_min_r],
                [-hd_center, -x_min_r],
                [-center_r, -center_r],
                [-one_half, x_min_r],
                [-hd_center, x_min_r],
                [-center_r, center_r],
            ]
        ) + np.array([one_half, one_half])

        x_max_ctps = np.array(
            [
                [center_r, -center_r],
                [hd_center, -x_max_r],
                [one_half, -x_max_r],
                [center_r, center_r],
                [hd_center, x_max_r],
                [one_half, x_max_r],
            ]
        ) + np.array([one_half, one_half])

        spline_list.append(
            base.Bezier(degrees=[1, 1], control_points=center_points)
        )

        spline_list.append(
            base.Bezier(degrees=[2, 1], control_points=x_min_ctps)
        )

        spline_list.append(
            base.Bezier(degrees=[2, 1], control_points=x_max_ctps)
        )

        spline_list.append(
            base.Bezier(degrees=[1, 2], control_points=y_min_ctps)
        )

        spline_list.append(
            base.Bezier(degrees=[1, 2], control_points=y_max_ctps)
        )

        return spline_list



"""gustaf/spline/visualize.py.

Spline visualization module. Supports visualization of following spline
(para_dim, dim) pairs: (1, 2), (1, 3), (2, 2), (2, 3), (3, 3)
"""
import numpy as np

from gustaf import Vertices, settings
from gustaf.helpers import options
from gustaf.utils import log
from gustaf.utils.arr import enforce_len


class SplineShowOption(options.ShowOption):
    """
    Show options for splines.
    """

    __slots__ = ()

    # if we start to support more backends, most of this options should become
    # some sort of spline common.
    _valid_options = options.make_valid_options(
        *options.vedo_common_options,
        options.Option(
            "vedo",
            "control_points",
            "Show spline's control points."
            "Options propagates to control mesh, unless it is specified.",
            (bool,),
        ),
        options.Option(
            "vedo",
            "control_mesh",
            "Show spline's control mesh.",
            (bool,),
        ),
        options.Option("vedo", "knots", "Show spline's knots.", (bool,)),
        options.Option(
            "vedo",
            "fitting_queries",
            "Shows fitting queries if they are locally saved in splines.",
            (bool,),
        ),
        options.Option(
            "vedo",
            "control_points_alpha",
            "Transparency of control points in range [0, 1].",
            (float, int),
        ),
        options.Option(
            "vedo",
            "control_point_ids",
            "Show ids of control_points.",
            (bool,),
        ),
        options.Option(
            "vedo",
            "resolutions",
            "Sampling resolution for spline.",
            (int, list, tuple, np.ndarray),
        ),
        options.Option(
            "vedo",
            "arrow_data_on",
            "Specify parametric coordinates to place arrow_data.",
            (list, tuple, np.ndarray),
        ),
    )

    _helps = "GustafSpline"

    def __init__(self, helpee):
        """
        Parameters
        ----------
        helpee: GustafSpline
        """
        self._helpee = helpee
        # checks if helpee inherits from GustafSpline
        if self._helps not in str(type(helpee).__mro__):
            raise TypeError(
                f"This show option if for {self._helps}.",
                f"Given helpee is {type(helpee)}.",
            )
        self._options = dict()
        self._backend = settings.VISUALIZATION_BACKEND
        self._options[self._backend] = dict()


def make_showable(spline):
    return eval(f"_{spline.show_options._backend}_showable(spline)")


def _vedo_showable(spline):
    """
    Goes through common procedures for preparing showable splines.

    Parameters
    ----------
    None

    Returns
    -------
    gus_dict: dict
      dict of sampled spline as gustaf objects
    """
    # get spline and knots
    gus_primitives = eval(f"_vedo_showable_para_dim_{spline.para_dim}(spline)")

    # apply spline color
    sampled_spline = gus_primitives["spline"]
    default_color = "green" if spline.para_dim > 1 else "black"
    sampled_spline.show_options["c"] = spline.show_options.get(
        "c", default_color
    )
    sampled_spline.show_options["alpha"] = spline.show_options.get("alpha", 1)
    sampled_spline.show_options["lighting"] = spline.show_options.get(
        "lighting", "glossy"
    )

    # axis?
    sampled_spline.show_options["axes"] = spline.show_options.get(
        "axes", False
    )

    # with color representable, scalar field data
    res = enforce_len(
        spline.show_options.get("resolutions", 100), spline.para_dim
    )
    data_name = spline.show_options.get("data_name", None)
    sampled_spline_data = spline.spline_data.as_scalar(
        data_name, res, default=None
    )
    if data_name is not None and sampled_spline_data is not None:
        # transfer data
        sampled_spline.vertex_data[data_name] = sampled_spline_data

        # transfer options - maybe vectorized query?
        keys = ("vmin", "vmax", "scalarbar", "cmap", "cmap_alpha")
        spline.show_options.copy_valid_options(
            sampled_spline.show_options, keys
        )

        # mark that we want to see this data
        sampled_spline.show_options["data_name"] = data_name

    elif data_name is not None and sampled_spline_data is None:
        log.debug(f"No spline_data named ({data_name}) for {spline}. Skipping")

    # with arrow representable, vector data
    adata_name = spline.show_options.get("arrow_data", None)
    adapted_adata = spline.spline_data.get(adata_name, None)
    if adata_name is not None and adapted_adata is not None:
        # if location is specified, this will be a separate Vertices obj with
        # configured arrow_data
        has_locations = adapted_adata.has_locations
        adata_on = "arrow_data_on" in spline.show_options.keys()
        create_vertices = has_locations or adata_on

        # this case causes conflict of interest. raise
        if has_locations and adata_on:
            raise ValueError(
                f"arrow_data-({adata_name}) has fixed location, "
                "but and `arrow_data_on` is set.",
            )

        if create_vertices:
            # prepare corresponding queries
            if adapted_adata.has_locations:
                queries = adapted_adata.locations
                on = None
            else:
                queries = spline.show_options["arrow_data_on"]
                on = queries

            # bound /  dim check
            bounds = spline.parametric_bounds
            if queries.shape[1] != len(bounds[0]):
                raise ValueError(
                    "Dimension mismatch: arrow_data locations-"
                    f"{queries.shape[1]} / para_dim-{spline.para_dim}."
                )
            # tolerance padding. may still cause issues in splinepy.
            # in that case, we will have to scale queries.
            lb_diff = queries.min(axis=0) - bounds[0] + settings.TOLERANCE
            ub_diff = queries.max(axis=0) - bounds[1] - settings.TOLERANCE
            if any(lb_diff < 0) or any(ub_diff > 0):
                raise ValueError(
                    f"Specified locations of ({adata_name}) are out side the "
                    f"parametric bounds ({bounds}) by [{lb_diff}, {ub_diff}]."
                )

            # get arrow
            adata = spline.spline_data.as_arrow(adata_name, on=on)

            # create vertices that can be shown as arrows
            loc_vertices = Vertices(spline.evaluate(queries), copy=False)
            loc_vertices.vertex_data[adata_name] = adata

            # transfer options
            keys = ("arrow_data_scale", "arrow_data_color", "arrow_data")
            spline.show_options.copy_valid_options(
                loc_vertices.show_options, keys
            )

            # add to primitives
            gus_primitives["arrow_data"] = loc_vertices

        else:  # sample arrows and append to sampled spline.
            sampled_spline.vertex_data[
                adata_name
            ] = spline.spline_data.as_arrow(adata_name, resolutions=res)
            # transfer options
            keys = ("arrow_data_scale", "arrow_data_color", "arrow_data")
            spline.show_options.copy_valid_options(
                sampled_spline.show_options, keys
            )

    # double check on same obj ref
    gus_primitives["spline"] = sampled_spline

    # control_points & control_points_alpha
    show_cps = spline.show_options.get("control_points", True)
    if show_cps:
        # control points (vertices)
        cps = spline.extract.control_points()
        cps.show_options["c"] = "red"
        cps.show_options["r"] = 10
        cps.show_options["alpha"] = spline.show_options.get(
            "control_points_alpha", 0.8
        )
        # add
        gus_primitives["control_points"] = cps

        if spline.show_options.get("control_point_ids", True):
            # a bit redundant, but nicely separable
            cp_ids = spline.extract.control_points()
            cp_ids.show_options["labels"] = np.arange(len(cp_ids.vertices))
            cp_ids.show_options["label_options"] = {"font": "VTK"}
            gus_primitives["control_point_ids"] = cp_ids

    if spline.show_options.get("control_mesh", show_cps):
        # pure control mesh
        c_mesh = spline.extract.control_mesh()  # either edges or faces
        if spline.para_dim != 1:
            c_mesh = c_mesh.to_edges(unique=True)

        c_mesh.show_options["c"] = "red"
        c_mesh.show_options["lw"] = 4
        c_mesh.show_options["alpha"] = spline.show_options.get(
            "control_points_alpha", 0.8
        )
        # add
        gus_primitives["control_mesh"] = c_mesh

    # fitting queries
    if hasattr(spline, "_fitting_queries") and spline.show_options.get(
        "fitting_queries", True
    ):
        fqs = Vertices(spline._fitting_queries)
        fqs.show_options["c"] = "blue"
        fqs.show_options["r"] = 10
        gus_primitives["fitting_queries"] = fqs

    return gus_primitives


def _vedo_showable_para_dim_1(spline):
    """Assumes showability check has been already performed.

    Parameters
    ----------
    spline: GustafSpline

    Returns
    -------
    gus_primitives: dict
      keys are {spline, knots}
    """
    gus_primitives = dict()
    res = enforce_len(
        spline.show_options.get("resolutions", 100), spline.para_dim
    )
    sp = spline.extract.edges(res[0])

    # specify curve width
    sp.show_options["lw"] = 8
    # add spline
    gus_primitives["spline"] = sp

    # place knots
    if spline.show_options.get("knots", True):
        uks = np.asanyarray(spline.unique_knots[0]).reshape(-1, 1)
        knots = Vertices(spline.evaluate(uks))
        knots.show_options["labels"] = ["x"] * len(uks)
        knots.show_options["label_options"] = {
            "justify": "center",
            "c": "green",
        }
        gus_primitives["knots"] = knots

    return gus_primitives


def _vedo_showable_para_dim_2(spline):
    """
    Assumes showability check has been already performed

    Parameters
    ----------
    spline: GustafSpline

    Returns
    -------
    gus_primitives: dict
      keys are {spline, knots}
    """
    gus_primitives = dict()
    res = enforce_len(
        spline.show_options.get("resolutions", 100), spline.para_dim
    )
    sp = spline.extract.faces(res)
    gus_primitives["spline"] = sp

    # knots
    if spline.show_options.get("knots", True):
        knot_lines = spline.extract.edges(res, all_knots=True)
        knot_lines.show_options["c"] = "black"
        knot_lines.show_options["lw"] = 3
        gus_primitives["knots"] = knot_lines

    return gus_primitives


def _vedo_showable_para_dim_3(spline):
    """
    Currently same as _vedo_showable_para_dim_2
    """
    return _vedo_showable_para_dim_2(spline)



import sys
from typing import Any, Union

from gustaf.edges import Edges
from gustaf.faces import Faces
from gustaf.vertices import Vertices
from gustaf.volumes import Volumes

MESH_TYPES = Union[Vertices, Edges, Faces, Volumes]


def is_mesh(candidate: Any) -> bool:
    """This function checks if the candidate is a mesh.

    The complicated version check is so that even python 3.6 is supported.

    Parameters
    -----------
    candidate: Any
      object to check for being a mesh.

    Returns
    --------
    is_mesh: bool
      Is the given object a mesh.
    """
    if float(sys.version.split(".")[1]) > 6:
        return isinstance(candidate, MESH_TYPES.__args__)
    else:
        return issubclass(type(candidate), MESH_TYPES)
